from matplotlib import pyplot as plt
import numpy as np


def soft_l1(z):
    return 2 * ((1 + z) ** 0.5 - 1)


x = np.linspace(-3, 3)
z = x**2
plt.plot(x, z, label="linear $\\rho(z) = z$")
plt.plot(x, soft_l1(z), label="soft L1-norm $\\rho(z) = 2(\\sqrt{1+z} - 1)$")
plt.xlabel("studentized residual")
plt.ylabel("cost")
plt.legend()



from iminuit import cost
from iminuit._hide_modules import hide_modules
import pytest

pytest.importorskip("matplotlib.pyplot")


def test_visualize():
    import iminuit

    c = cost.LeastSquares([1, 2], [3, 4], 1, lambda x, a: a * x)

    s = iminuit.Minuit(c, 1).migrad()._repr_html_()
    assert "<svg" in s

    with hide_modules("matplotlib", reload="iminuit"):
        s = iminuit.Minuit(c, 1).migrad()._repr_html_()
        assert "<svg" not in s



"""Print iminuit version."""

from pathlib import PurePath

try:
    import tomllib as toml
except ModuleNotFoundError:
    import tomli as toml

with open(PurePath(__file__).parent / "pyproject.toml", "rb") as f:
    version = toml.load(f)["project"]["version"]

if __name__ == "__main__":
    print(version)



"""
Types for iminuit.

These are used by mypy and similar tools.
"""

from typing import (
    Protocol,
    Optional,
    List,
    Union,
    runtime_checkable,
    NamedTuple,
    Annotated,
)
from numpy.typing import NDArray
import numpy as np
import dataclasses


# Key for ValueView, ErrorView, etc.
Key = Union[int, str, slice, List[Union[int, str]]]


@runtime_checkable
class Model(Protocol):
    """Type for user-defined model."""

    def __call__(self, x: np.ndarray, *args: float) -> np.ndarray:
        """Evaluate model at locations x and return results as an array."""
        ...  # pragma: no cover


@runtime_checkable
class ModelGradient(Protocol):
    """Type for user-defined model gradient."""

    def __call__(self, x: np.ndarray, *args: float) -> np.ndarray:
        """Evaluate model gradient at locations x and return results as an array."""
        ...  # pragma: no cover


@runtime_checkable
class Cost(Protocol):
    """Type for user-defined cost function."""

    def __call__(self, *args: float) -> float:
        """Evaluate cost and return results as a float."""
        ...  # pragma: no cover


@runtime_checkable
class CostGradient(Protocol):
    """Type for user-defined gradient of a cost function."""

    def __call__(self, *args: float) -> np.ndarray:
        """Evaluate gradient and return results as an array."""
        ...  # pragma: no cover


@runtime_checkable
class LossFunction(Protocol):
    """Type for user-defined loss function for LeastSquares clas."""

    def __call__(self, z: NDArray) -> NDArray:
        """Evaluate loss function on values."""
        ...  # pragma: no cover


class UserBound(NamedTuple):
    """Type for user-defined limit."""

    min: Optional[float]
    max: Optional[float]


@dataclasses.dataclass
class Gt:
    """Annotation compatible with annotated-types."""

    gt: float


@dataclasses.dataclass
class Ge:
    """Annotation compatible with annotated-types."""

    ge: float


@dataclasses.dataclass
class Lt:
    """Annotation compatible with annotated-types."""

    lt: float


@dataclasses.dataclass
class Le:
    """Annotation compatible with annotated-types."""

    le: float


@dataclasses.dataclass
class Interval:
    """Annotation compatible with annotated-types."""

    gt: Optional[float] = None
    ge: Optional[float] = None
    lt: Optional[float] = None
    le: Optional[float] = None


# common convenience types
PositiveFloat = Annotated[float, Gt(0)]
Probability = Annotated[float, Interval(ge=0, le=1)]



from iminuit.cost import UnbinnedNLL
from iminuit import Minuit
import numpy as np
import numba as nb
from numba_stats import norm, truncexpon
import pytest
from numpy.testing import assert_allclose

N = [int(x) for x in np.geomspace(10, 1e6, 11)]
# N = N[:-5]


def make_data(size, seed=1):
    rng = np.random.default_rng(seed)
    s = rng.normal(0.5, 0.1, size=size // 2)
    b = rng.exponential(1, size=2 * size)
    b = b[b < 1]
    b = b[: size // 2]
    x = np.append(s, b)
    return x


# we inline following functions to profit from parallel=True,fastmath=True


@nb.njit(inline="always")
def mixture(x, z, mu, sigma, slope):
    b = truncexpon.pdf(x, 0.0, 1.0, 0.0, slope)
    s = norm.pdf(x, mu, sigma)
    return (1 - z) * b + z * s


@nb.njit(inline="always")
def logsumexp(a, b):
    r = np.empty_like(a)
    for i in nb.prange(len(r)):
        if a[i] > b[i]:
            c = a[i]
            d = b[i]
        else:
            c = b[i]
            d = a[i]
        r[i] = c + np.log1p(np.exp(d - c))
    return r


ARGS = (0.5, 0.5, 0.1, 1.0)


@nb.njit(inline="always")
def log_mixture(x, z, mu, sigma, slope):
    log_b = truncexpon.logpdf(x, 0.0, 1.0, 0.0, slope)
    log_s = norm.logpdf(x, mu, sigma)
    b = np.log(1 - z) + log_b
    s = np.log(z) + log_s
    return logsumexp(b, s)


@pytest.mark.parametrize("n", N)
@pytest.mark.parametrize("log", [False, True])
def test_UnbinnedNLL(benchmark, n, log):
    x = make_data(size=n)
    fn = log_mixture if log else mixture
    fn(x, *ARGS)  # warm-up JIT
    cost = UnbinnedNLL(x, fn, log=log)
    benchmark(cost, *ARGS)


@pytest.mark.parametrize("n", N)
@pytest.mark.parametrize("lib", ["numba_stats", "scipy.stats"])
def test_nll(benchmark, n, lib):
    x = make_data(size=n)

    if lib == "scipy.stats":
        from scipy.stats import truncexpon, norm

        def fn(x, z, mu, sigma, slope):
            b = truncexpon.pdf(x, 1.0, 0.0, slope)
            s = norm.pdf(x, mu, sigma)
            return (1 - z) * b + z * s

    else:
        fn = mixture

    def cost(x, z, mu, sigma, slope):
        return -np.sum(np.log(fn(x, z, mu, sigma, slope)))

    benchmark(cost, x, *ARGS)


@pytest.mark.parametrize("n", N)
def test_nll_numba(benchmark, n):
    x = make_data(size=n)

    @nb.njit
    def cost(x, z, mu, sigma, slope):
        return -np.sum(np.log(mixture(x, z, mu, sigma, slope)))

    benchmark(cost, x, *ARGS)


@pytest.mark.parametrize("n", N)
@pytest.mark.parametrize("numba", [False, True])
@pytest.mark.parametrize("log", [False, True])
def test_minuit(benchmark, n, log, numba):
    x = make_data(size=n)

    if log:

        def cost(x, z, mu, sigma, slope):
            return -np.sum(log_mixture(x, z, mu, sigma, slope))

    else:

        def cost(x, z, mu, sigma, slope):
            return -np.sum(np.log(mixture(x, z, mu, sigma, slope)))

    if numba:
        cost = nb.njit(cost)
        cost(x, *ARGS)  # jit warm-up

    m = Minuit(lambda *args: cost(x, *args), *ARGS)
    m.errordef = Minuit.LIKELIHOOD
    m.limits[0, 1] = (0, 1)
    m.limits[2, 3] = (0, 10)

    def run():
        m.reset()
        return m.migrad()

    m = benchmark(run)
    assert m.valid
    # ignore slope
    assert_allclose(m.values[:-1], ARGS[:-1], atol=2 / n**0.5)


@pytest.mark.parametrize("n", N)
@pytest.mark.parametrize("log", [False, True])
def test_minuit_parallel_fastmath(benchmark, n, log):
    x = make_data(size=n)

    if log:

        @nb.njit(parallel=True, fastmath=True)
        def cost(x, z, mu, sigma, slope):
            return -np.sum(log_mixture(x, z, mu, sigma, slope))

    else:

        @nb.njit(parallel=True, fastmath=True)
        def cost(x, z, mu, sigma, slope):
            p = mixture(x, z, mu, sigma, slope)
            return -np.sum(np.log(p))

    cost(x, *ARGS)  # jit warm-up

    m = Minuit(lambda *args: cost(x, *args), *ARGS)
    m.errordef = Minuit.LIKELIHOOD
    m.limits[0, 1] = (0, 1)
    m.limits[2, 3] = (0, 10)

    def run():
        m.reset()
        return m.migrad()

    m = benchmark(run)
    assert m.valid
    # ignore slope
    assert_allclose(m.values[:-1], ARGS[:-1], atol=2 / n**0.5)


@pytest.mark.parametrize("n", N)
def test_minuit_cfunc(benchmark, n):
    x = make_data(size=n)

    @nb.cfunc(nb.double(nb.uintc, nb.types.CPointer(nb.double)))
    def cost(n, par):
        z, mu, sigma, slope = nb.carray(par, (n,))
        return -np.sum(np.log(mixture(x, z, mu, sigma, slope)))

    m = Minuit(cost, *ARGS)
    m.errordef = Minuit.LIKELIHOOD
    m.limits[0, 1] = (0, 1)
    m.limits[2, 3] = (0, 10)

    def run():
        m.reset()
        return m.migrad()

    m = benchmark(run)
    assert m.valid
    # ignore slope
    assert_allclose(m.values[:-1], ARGS[:-1], atol=2 / n**0.5)


@pytest.mark.parametrize("n", N)
@pytest.mark.parametrize("log", [False, True])
def test_minuit_UnbinnedNLL(benchmark, n, log):
    x = make_data(size=n)

    cost = UnbinnedNLL(x, log_mixture if log else mixture, log=log)
    m = Minuit(cost, *ARGS)
    m.limits[0, 1] = (0, 1)
    m.limits[2, 3] = (0, 10)

    def run():
        m.reset()
        return m.migrad()

    m = benchmark(run)
    assert m.valid
    # ignore slope
    assert_allclose(m.values[:-1], ARGS[:-1], atol=2 / n**0.5)


try:
    import ROOT as R

    if R.__version__ >= "6.30":

        @pytest.mark.parametrize("n", N)
        @pytest.mark.parametrize(
            "backend",
            [
                "legacy",
                "legacy-parallel",
                "cpu",
                "cpu-parallel",
                "cpu-implicitmt",
                "codegen",
                "codegen_no_grad",
            ],
        )
        def test_RooFit(benchmark, n, backend):
            x = R.RooRealVar("x", "x", 0, 1)
            z = R.RooRealVar("z", "z", 0.5, 0, 1)
            mu = R.RooRealVar("mu", "mu", 0.5, 0, 1)
            sigma = R.RooRealVar("sigma", "sigma", 0.1, 0, 10)
            slope = R.RooRealVar("slope", "slope", 1.0, 0, 10)
            pdf1 = R.RooGaussian("gauss", "gauss", x, mu, sigma)
            pdf2 = R.RooExponential("expon", "expon", x, slope)
            pdf = R.RooAddPdf("pdf", "pdf", [pdf1, pdf2], [z])

            data = pdf.generate(x, n)

            parallel = False
            implicitmt = False
            if backend.endswith("-parallel"):
                backend = backend.split("-")[0]
                parallel = True
            if backend.endswith("-implicitmt"):
                backend = backend.split("-")[0]
                implicitmt = True
            print(backend)
            args = [R.RooFit.PrintLevel(-1), R.RooFit.EvalBackend(backend)]

            NumCPU = nb.get_num_threads()
            if parallel:
                args.append(R.RooFit.NumCPU(NumCPU))
            if implicitmt:
                R.EnableImplicitMT(NumCPU)
            else:
                R.EnableImplicitMT(1)

            def run():
                mu.setVal(0.5)
                sigma.setVal(0.1)
                slope.setVal(1)
                z.setVal(0.5)
                pdf.fitTo(data, *args)

            benchmark(run)
            assert_allclose(z.getVal(), 0.5, atol=5 / n**0.5)
            assert_allclose(mu.getVal(), 0.5, atol=5 / n**0.5)
            assert_allclose(sigma.getVal(), 0.1, atol=5 / n**0.5)

    else:

        @pytest.mark.parametrize("n", N)
        @pytest.mark.parametrize("NumCPU", [0, nb.get_num_threads()])
        @pytest.mark.parametrize("BatchMode", [False, True])
        def test_RooFit(benchmark, n, NumCPU, BatchMode):
            x = R.RooRealVar("x", "x", 0, 1)
            z = R.RooRealVar("z", "z", 0.5, 0, 1)
            mu = R.RooRealVar("mu", "mu", 0.5, 0, 1)
            sigma = R.RooRealVar("sigma", "sigma", 0.1, 0, 10)
            slope = R.RooRealVar("slope", "slope", 1.0, 0, 10)
            pdf1 = R.RooGaussian("gauss", "gauss", x, mu, sigma)
            pdf2 = R.RooExponential("expon", "expon", x, slope)
            pdf = R.RooAddPdf("pdf", "pdf", [pdf1, pdf2], [z])

            data = pdf.generate(x, n)

            args = [R.RooFit.PrintLevel(-1), R.RooFit.BatchMode(BatchMode)]
            if NumCPU:
                args.append(R.RooFit.NumCPU(NumCPU, 1))
                args.append(R.RooFit.Parallelize(NumCPU))

            def run():
                mu.setVal(0.5)
                sigma.setVal(0.1)
                slope.setVal(1)
                z.setVal(0.5)
                pdf.fitTo(data, *args)

            benchmark(run)
            assert_allclose(z.getVal(), 0.5, atol=5 / n**0.5)
            assert_allclose(mu.getVal(), 0.5, atol=5 / n**0.5)
            assert_allclose(sigma.getVal(), 0.1, atol=5 / n**0.5)

except ModuleNotFoundError:
    pass



"""Ensure that current version is not in conflict with published releases."""

from packaging.version import Version
import subprocess as subp
from pathlib import PurePath
import urllib.request
import json
import warnings
import sys


project_dir = PurePath(__file__).parent.parent
changelog_fn = project_dir / "doc/changelog.rst"
version_fn = project_dir / "version.py"

version = subp.check_output([sys.executable, version_fn]).strip().decode()

with warnings.catch_warnings(record=True) as record:
    iminuit_version = Version(version)
if record:
    raise ValueError(record[0].message)

print("iminuit version:", iminuit_version)

# make sure that changelog was updated
with open(changelog_fn) as f:
    assert iminuit_version.base_version in f.read(), "changelog entry missing"

# make sure that version is not already tagged
tags = subp.check_output(["git", "tag"]).decode().strip().split("\n")
assert f"v{iminuit_version}" not in tags, "tag exists"

# make sure that version itself was updated
with urllib.request.urlopen("https://pypi.org/pypi/iminuit/json") as r:
    pypi_versions = [Version(v) for v in json.loads(r.read())["releases"]]

pypi_versions.sort()
print("PyPI    version:", pypi_versions[-1])

assert iminuit_version not in pypi_versions, "pypi version exists"



from iminuit._hide_modules import hide_modules
import numpy as np
import pytest
from numpy.testing import assert_allclose

pytest.importorskip("numba")


def npa(*args, **kwargs):
    return np.array(args, **kwargs)


@pytest.mark.parametrize(
    "func,args",
    (
        ("log_or_zero", (npa(1.2, 0),)),
        ("_z_squared", (npa(1.2, 0), npa(1.2, 1.0), npa(1.2, 1))),
        ("_unbinned_nll", (npa(1.2, 0),)),
        ("multinomial_chi2", (npa(1, 0), npa(1.2, 0))),
        ("chi2", (npa(1.2, 0), npa(1.2, 1.0), npa(1.2, 0.1))),
        ("poisson_chi2", (npa(1, 0), npa(1.2, 0.1))),
        ("_soft_l1_cost", (npa(1.2, 0), npa(1.2, 0.1), npa(1.0, 1.0))),
    ),
)
def test_no_numba(func, args):
    import iminuit.cost as cost

    assert "jit" in dir(cost)

    expected = getattr(cost, func)(*args)

    with hide_modules("numba", reload="iminuit.cost"):
        import iminuit.cost as cost

        assert "jit" not in dir(cost)

        got = getattr(cost, func)(*args)

    assert_allclose(got, expected)



import numpy as np
import warnings


def test_issue_424():
    from iminuit import Minuit

    def fcn(x, y, z):
        return (x - 1) ** 2 + (y - 4) ** 2 / 2 + (z - 9) ** 2 / 3

    m = Minuit(fcn, x=0.0, y=0.0, z=0.0)
    m.migrad()

    m.fixed["x"] = True
    m.errors["x"] = 2
    m.hesse()  # this used to release x
    assert m.fixed["x"]
    assert m.errors["x"] == 2


def test_issue_544():
    import pytest
    from iminuit import Minuit
    from iminuit.util import IMinuitWarning

    def fcn(x, y):
        return x**2 + y**2

    m = Minuit(fcn, x=0, y=0)
    m.fixed = True
    with pytest.warns(IMinuitWarning):
        m.hesse()  # this used to cause a segfault


def test_issue_648():
    from iminuit import Minuit

    class F:
        first = True

        def __call__(self, a, b):
            if self.first:
                assert a == 1.0 and b == 2.0
                self.first = False
            return a**2 + b**2

    m = Minuit(F(), a=1, b=2)
    m.fixed["a"] = False  # this used to change a to b
    m.migrad()


def test_issue_643():
    from iminuit import Minuit

    def fcn(x, y, z):
        return (x - 2) ** 2 + (y - 3) ** 2 + (z - 4) ** 2

    m = Minuit(fcn, x=2, y=3, z=4)
    m.migrad()

    m2 = Minuit(fcn, x=m.values["x"], y=m.values["y"], z=m.values["z"])
    # used to call MnHesse when it was not needed and quickly exhaust call limit
    for i in range(10):
        m2.minos()

    m2.reset()
    # used to exhaust call limit, because calls to MnHesse did not reset call count
    for i in range(10):
        m2.values = m.values
        m2.minos()


def test_issue_669():
    from iminuit import Minuit

    def fcn(x, y):
        return x**2 + (y / 2) ** 2

    m = Minuit(fcn, x=0, y=0)

    m.migrad()

    xy1 = m.mncontour(x="x", y="y", size=10)
    xy2 = m.mncontour(x="y", y="x", size=10)  # used to fail

    # needs better way to compare polygons
    for x, y in xy1:
        match = False
        for y2, x2 in xy2:
            if abs(x - x2) < 1e-3 and abs(y - y2) < 1e-3:
                match = True
                break
        assert match


# cannot define this inside function, pickle will not allow it
def fcn(par):
    return np.sum(par**2)


# cannot define this inside function, pickle will not allow it
def grad(par):
    return 2 * par


def test_issue_687():
    import pickle
    import numpy as np
    from iminuit import Minuit

    start = np.zeros(3)
    m = Minuit(fcn, start)

    m.migrad()
    s_m = str(m)

    s = pickle.dumps(m)
    m2 = pickle.loads(s)

    s_m2 = str(m2)  # this used to fail
    assert s_m == s_m2


def test_issue_694():
    import pytest
    import numpy as np
    from iminuit import Minuit
    from iminuit.cost import ExtendedUnbinnedNLL

    stats = pytest.importorskip("scipy.stats")

    xmus = 1.0
    xmub = 5.0
    xsigma = 1.0
    ymu = 0.5
    ysigma = 0.2
    ytau = 0.1

    for seed in range(100):
        rng = np.random.default_rng(seed)

        xs = rng.normal(xmus, xsigma, size=33)
        xb = rng.normal(xmub, xsigma, size=66)
        x = np.append(xs, xb)

        def model(x, sig_n, sig_mu, sig_sigma, bkg_n, bkg_tau):
            return sig_n + bkg_n, (
                sig_n * stats.norm.pdf(x, sig_mu, sig_sigma)
                + bkg_n * stats.expon.pdf(x, 0, bkg_tau)
            )

        nll = ExtendedUnbinnedNLL(x, model)

        m = Minuit(nll, sig_n=33, sig_mu=ymu, sig_sigma=ysigma, bkg_n=66, bkg_tau=ytau)
        # with Simplex the fit never yields NaN, which is good but not what we want here
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", RuntimeWarning)
            m.migrad(use_simplex=False)

        if np.isnan(m.fmin.edm):
            assert not m.valid
            assert m.fmin.is_above_max_edm
            break
    else:
        assert False


def test_issue_923():
    from iminuit import Minuit
    from iminuit.cost import LeastSquares
    import numpy as np
    import pytest

    # implicitly needed by visualize
    pytest.importorskip("matplotlib")

    def model(x, c1):
        c2 = 100
        res = np.zeros(len(x))
        mask = x < 47
        res[mask] = c1
        res[~mask] = c2
        return res

    xtest = np.linspace(0, 74)
    ytest = xtest * 0 + 1
    ytesterr = ytest

    least_squares = LeastSquares(xtest, ytest, ytesterr, model)

    m = Minuit(least_squares, c1=1)
    m.migrad()
    # this used to trigger an endless (?) loop
    m.visualize()



"""Get the latest Python release which is online."""

import urllib.request
import re
from html.parser import HTMLParser
import gzip


class PythonVersionParser(HTMLParser):
    """Specialized HTMLParser to get Python version number."""

    def __init__(self):
        super().__init__()
        self.versions = set()
        self.found_version = False

    def handle_starttag(self, tag, attrs):
        """Look for the right tag and store result in an attribute."""
        if tag == "a":
            for attr in attrs:
                if attr[0] == "href" and "/downloads/release/python-" in attr[1]:
                    self.found_version = True
                    return

    def handle_data(self, data):
        """Extract Python version from entry."""
        if self.found_version:
            self.found_version = False
            match = re.search(r"Python (\d+)\.(\d+)\.(\d+)?", data)
            if match:
                major = int(match.group(1))
                minor = int(match.group(2))
                bugfix = int(match.group(3))
                self.versions.add((major, minor, bugfix))


def versions():
    """Get all Python release versions."""
    req = urllib.request.Request("https://www.python.org/downloads/")
    req.add_header("Accept-Encoding", "gzip")

    with urllib.request.urlopen(req) as response:
        raw = response.read()
        if response.info().get("Content-Encoding") == "gzip":
            raw = gzip.decompress(raw)
        html = raw.decode("utf-8")

    parser = PythonVersionParser()
    parser.feed(html)

    return parser.versions


def latest():
    """Return version of latest Python release."""
    return max(versions())


def main():
    """Print all discovered release versions."""
    for x in sorted(versions()):
        print(x)


if __name__ == "__main__":
    main()



from pathlib import Path
import sys


def install(packages):
    import subprocess as subp

    subp.check_call(
        ["python", "-m", "pip", "install", "--upgrade", "--prefer-binary"] + packages
    )


packages = [
    "tomli",
    "pip",
    "wheel",
]

install(packages)

import tomli  # noqa

with open(Path(__file__).parent.parent / "pyproject.toml", "rb") as f:
    d = tomli.load(f)

for arg in sys.argv[1:]:
    if arg == "build":
        install(d["build-system"]["requires"])
    else:
        packages = d["project"]["optional-dependencies"][arg]
        install(packages)



from iminuit import Minuit


def cost(x, y, z):
    return (x - 1) ** 2 + (y - x) ** 2 + (z - 2) ** 2


cost.errordef = Minuit.LEAST_SQUARES

m = Minuit(cost, x=0, y=0, z=0)
m.migrad()
m.draw_mnprofile("y")



import subprocess as subp
from pathlib import PurePath

project_dir = PurePath(__file__).parent.parent

# check that root version is up-to-date
git_submodule = subp.check_output(
    ["git", "submodule", "status"], cwd=project_dir
).decode()

for item in git_submodule.strip().split("\n"):
    parts = item.split()
    if PurePath(parts[1]) != PurePath("extern") / "root":
        continue

    assert len(parts) == 3, "module is not checked out"
    break

# git submodule status does not yield the right state
# we must use git describe --tags
root_version = (
    subp.check_output(
        ["git", "describe", "--tags"], cwd=project_dir / "extern" / "root"
    )
    .decode()
    .strip()
)
print("ROOT", root_version)



from __future__ import annotations
import pytest
import numpy as np
from numpy.testing import assert_allclose, assert_equal
from iminuit import Minuit
from iminuit.cost import (
    CostSum,
    UnbinnedNLL,
    BinnedNLL,
    ExtendedUnbinnedNLL,
    ExtendedBinnedNLL,
    LeastSquares,
    Constant,
    NormalConstraint,
    Template,
    multinomial_chi2,
    PerformanceWarning,
)
from iminuit.util import describe
from iminuit.typing import Annotated, Gt, Lt
from typing import Sequence
import pickle


def norm_logpdf(x, mu, sigma):
    z = (x - mu) / sigma
    return -0.5 * (np.log(2 * np.pi) + z**2) - np.log(sigma)


def norm_pdf(x, mu, sigma):
    return np.exp(norm_logpdf(x, mu, sigma))


def norm_cdf(x, mu, sigma):
    from math import erf

    z = (x - mu) / (np.sqrt(2) * sigma)

    return (1 + np.vectorize(erf)(z)) * 0.5


def mvnorm(mux, muy, sx, sy, rho):
    stats = pytest.importorskip("scipy.stats")
    C = np.empty((2, 2))
    C[0, 0] = sx**2
    C[0, 1] = C[1, 0] = sx * sy * rho
    C[1, 1] = sy**2
    m = [mux, muy]
    return stats.multivariate_normal(m, C)


def expon_cdf(x, a):
    with np.errstate(over="ignore"):
        return -np.expm1(-x / a)


def scaled_expon_cdf(x, n, a):
    with np.errstate(over="ignore"):
        return n * -np.expm1(-x / a)


def numerical_cost_gradient(fcn):
    jacobi = pytest.importorskip("jacobi").jacobi
    return lambda *args: jacobi(lambda p: fcn(*p), args)[0]


def numerical_model_gradient(fcn):
    jacobi = pytest.importorskip("jacobi").jacobi
    return lambda x, *args: jacobi(lambda p: fcn(x, *p), args)[0].T


def numerical_extended_model_gradient(fcn):
    jacobi = pytest.importorskip("jacobi").jacobi

    def fn(x, *args):
        fint = jacobi(lambda p: fcn(x, *p)[0], args)[0]
        f = jacobi(lambda p: fcn(x, *p)[1], args)[0].T
        return fint, f

    return fn


@pytest.fixture
def unbinned():
    rng = np.random.default_rng(1)
    x = rng.normal(size=1000)
    mle = (len(x), np.mean(x), np.std(x, ddof=1))
    return mle, x


@pytest.fixture
def binned(unbinned):
    mle, x = unbinned
    nx, xe = np.histogram(x, bins=50, range=(-3, 3))
    assert np.sum(nx == 0) > 0
    return mle, nx, xe


def logpdf(x, mu, sigma):
    return norm_logpdf(x, mu, sigma)


def pdf(x, mu, sigma):
    return norm_pdf(x, mu, sigma)


def pdf_nosig(x, *par):
    return norm_pdf(x, *par)


def annotated_pdf(x, mu, sigma: Annotated[float, 0 : np.inf]):
    return norm_pdf(x, mu, sigma)


def cdf(x, mu, sigma):
    return norm_cdf(x, mu, sigma)


def cdf_nosig(x, *par):
    return norm_cdf(x, *par)


def scaled_cdf(x, n, mu, sigma):
    return n * norm_cdf(x, mu, sigma)


def scaled_cdf_nosig(x, *par):
    return par[0] * norm_cdf(x, *par[1:])


def line(x, a, b):
    return a + b * x


def line_nosig(x, *par):
    return par[0] + par[1] * x


def test_norm_logpdf():
    norm = pytest.importorskip("scipy.stats").norm
    x = np.linspace(-3, 3)
    assert_allclose(norm_logpdf(x, 3, 2), norm.logpdf(x, 3, 2))


def test_norm_pdf():
    norm = pytest.importorskip("scipy.stats").norm
    x = np.linspace(-3, 3)
    assert_allclose(norm_pdf(x, 3, 2), norm.pdf(x, 3, 2))


def test_norm_cdf():
    norm = pytest.importorskip("scipy.stats").norm
    x = np.linspace(-3, 3)
    assert_allclose(norm_cdf(x, 3, 2), norm.cdf(x, 3, 2))


def test_describe():
    c = Constant(2.5)
    assert describe(c, annotations=True) == {}

    c = NormalConstraint("foo", 1.5, 0.1)
    assert describe(c, annotations=True) == {"foo": None}

    def model(
        x, foo: Annotated[float, Gt(1), Lt(2)], bar: float, baz: Annotated[float, 0:]
    ):
        return x

    assert describe(model, annotations=True) == {
        "x": None,
        "foo": (1, 2),
        "bar": None,
        "baz": (0, np.inf),
    }

    c = UnbinnedNLL([], model)
    assert describe(c, annotations=True) == {
        "foo": (1, 2),
        "bar": None,
        "baz": (0, np.inf),
    }

    c = BinnedNLL([], [1], model)
    assert describe(c, annotations=True) == {
        "foo": (1, 2),
        "bar": None,
        "baz": (0, np.inf),
    }


def test_Constant_1():
    c = Constant(2.5)
    assert c.value == 2.5
    assert c.ndata == 0
    assert c.has_grad
    assert_equal(c.grad(), [])


def test_Constant_2():
    c = NormalConstraint(("a", "b"), (1, 2), (3, 4)) + Constant(10.2)
    assert_allclose(c(1, 2), 10.2)
    assert_allclose(c(4, 2), 10.2 + 1)


def test_Constant_3():
    c = NormalConstraint(("a", "b"), (1, 2), (3, 4)) + Constant(10.2)
    ref = numerical_cost_gradient(c)
    assert c.has_grad
    assert_allclose(c.grad(1, 2), ref(1, 2))
    assert_allclose(c.grad(2, 3), ref(2, 3))
    assert_allclose(c.grad(-1, -2), ref(-1, -2))


@pytest.mark.parametrize("verbose", (0, 1))
@pytest.mark.parametrize("model", (logpdf, pdf))
@pytest.mark.parametrize("use_grad", (False, True))
def test_UnbinnedNLL(unbinned, verbose, model, use_grad):
    mle, x = unbinned

    cost = UnbinnedNLL(
        x,
        model,
        verbose=verbose,
        log=model is logpdf,
        grad=numerical_model_gradient(model),
    )
    assert cost.ndata == np.inf

    if use_grad:
        ref = numerical_cost_gradient(cost)
        assert_allclose(cost.grad(1, 2), ref(1, 2), rtol=1e-3)
        assert_allclose(cost.grad(-1, 3), ref(-1, 3), rtol=1e-3)

    m = Minuit(cost, mu=0, sigma=1, grad=use_grad)
    m.limits["sigma"] = (0, None)
    m.migrad()
    assert m.valid
    assert_allclose(m.values, mle[1:], atol=1e-3)
    assert m.errors["mu"] == pytest.approx(1000**-0.5, rel=0.05)

    if use_grad:
        assert m.ngrad > 0
    else:
        assert m.ngrad == 0

    assert_equal(m.fmin.reduced_chi2, np.nan)

    if use_grad and verbose == 0:
        fi = cost.fisher_information(*m.values)
        cov = cost.covariance(*m.values)
        assert_allclose(cov, np.linalg.inv(fi))
        assert_allclose(np.diag(cov) ** 0.5, m.errors, rtol=5e-2)
        rho1 = m.covariance.correlation()[0, 1]
        rho2 = cov[0, 1] / (cov[0, 0] * cov[1, 1]) ** 0.5
        assert_allclose(rho1, rho2, atol=0.01)


def test_UnbinnedNLL_name(unbinned):
    mle, x = unbinned

    cost = UnbinnedNLL(x, pdf_nosig, name=("mu", "sigma"))
    assert cost.ndata == np.inf

    m = Minuit(cost, mu=0, sigma=1)
    assert m.parameters == ("mu", "sigma")
    m.limits["sigma"] = (0, None)
    m.migrad()
    assert m.valid


def test_UnbinnedNLL_rename(unbinned):
    mle, x = unbinned

    cost = UnbinnedNLL(x, annotated_pdf, name=("m", "s"))
    m = Minuit(cost, m=0, s=1)
    assert m.limits["m"] == (-np.inf, np.inf)
    assert m.limits["s"] == (0, np.inf)
    assert m.parameters == ("m", "s")
    m.migrad()
    assert m.valid


def test_UnbinnedNLL_name_bad():
    with pytest.raises(
        ValueError, match="length of name does not match number of model parameters"
    ):
        UnbinnedNLL([1, 2, 3], pdf, name=("mu", "sigma", "foo"))


@pytest.mark.parametrize("use_grad", (False, True))
def test_UnbinnedNLL_2D(use_grad):
    def model(x_y, mux, muy, sx, sy, rho):
        return mvnorm(mux, muy, sx, sy, rho).pdf(x_y.T)

    truth = 0.1, 0.2, 0.3, 0.4, 0.5
    x, y = mvnorm(*truth).rvs(size=100, random_state=1).T

    cost = UnbinnedNLL((x, y), model, grad=numerical_model_gradient(model))
    m = Minuit(cost, *truth, grad=use_grad)
    m.limits["sx", "sy"] = (0, None)
    m.limits["rho"] = (-1, 1)
    m.migrad()
    assert m.valid

    if use_grad:
        assert m.ngrad > 0

        ref = numerical_cost_gradient(cost)
        assert_allclose(cost.grad(*m.values), ref(*m.values), rtol=0.01)
    else:
        assert m.ngrad == 0

    assert_allclose(m.values, truth, atol=0.2)


@pytest.mark.parametrize("use_grad", (False, True))
def test_UnbinnedNLL_mask(use_grad):
    c = UnbinnedNLL(
        [1, np.nan, 2],
        lambda x, a: x + a**2,
        grad=lambda x, a: np.ones_like(x) * 2 * a,
    )
    assert c.mask is None

    assert np.isnan(c(0))
    c.mask = np.arange(3) != 1
    assert_equal(c.mask, (True, False, True))
    assert not np.isnan(c(0))

    assert_allclose(c.grad(0), [0])

    ref = numerical_cost_gradient(c)
    assert_allclose(c.grad(1), ref(1), atol=1e-3)


@pytest.mark.parametrize("log", (False, True))
def test_UnbinnedNLL_properties(log):
    c = UnbinnedNLL([1, 2], norm_logpdf if log else norm_pdf, log=log)

    x = np.linspace(0, 1)
    expected = norm_pdf(x, 1, 2)
    assert_allclose(c.pdf(x, 1, 2), expected)

    expected *= len(c.data)
    assert_allclose(c.scaled_pdf(x, 1, 2), expected)

    with pytest.raises(AttributeError):
        c.pdf = None

    with pytest.raises(AttributeError):
        c.scaled_pdf = None

    assert_equal(c.data, [1, 2])
    c.data = [2, 3]
    assert_equal(c.data, [2, 3])
    with pytest.raises(ValueError):
        c.data = [1, 2, 3]
    assert c.verbose == 0
    c.verbose = 1
    assert c.verbose == 1


@pytest.mark.parametrize("log", (False, True))
def test_UnbinnedNLL_visualize(log):
    pytest.importorskip("matplotlib")
    c = UnbinnedNLL([1, 2], norm_logpdf if log else norm_pdf, log=log)
    # auto-sampling
    c.visualize((1, 2))
    # linear spacing
    c.visualize((1, 2), model_points=10)
    # linear spacing and different binning
    c.visualize((1, 2), model_points=10, bins=20)
    # trigger log-spacing
    c = UnbinnedNLL([1, 100, 1000], norm_logpdf if log else norm_pdf, log=log)
    c.visualize((1, 2), model_points=10)
    c.visualize((1, 2), model_points=10, bins=20)
    # manual spacing
    c.visualize((1, 2), model_points=np.linspace(1, 1000))

    with pytest.warns(FutureWarning, match="keyword 'nbins' is deprecated"):
        c.visualize((1, 2), nbins=20)


def test_UnbinnedNLL_visualize_2D():
    pytest.importorskip("matplotlib")

    def model(x_y, mux, muy, sx, sy, rho):
        return mvnorm(mux, muy, sx, sy, rho).pdf(x_y.T)

    truth = 0.1, 0.2, 0.3, 0.4, 0.5
    x, y = mvnorm(*truth).rvs(size=10, random_state=1).T

    c = UnbinnedNLL((x, y), model)

    with pytest.raises(ValueError, match="not implemented for multi-dimensional"):
        c.visualize(truth)


def test_UnbinnedNLL_pickle():
    c = UnbinnedNLL([1, 2], norm_pdf)
    b = pickle.dumps(c)
    c2 = pickle.loads(b)
    assert_equal(c.data, c2.data)


def test_UnbinnedNLL_annotated():
    c = UnbinnedNLL([], annotated_pdf)
    m = Minuit(c, mu=1, sigma=1.5)
    assert m.limits[0] == (-np.inf, np.inf)
    assert m.limits[1] == (0, np.inf)


@pytest.mark.parametrize("verbose", (0, 1))
@pytest.mark.parametrize("model", (logpdf, pdf))
@pytest.mark.parametrize("use_grad", (False, True))
def test_ExtendedUnbinnedNLL(unbinned, verbose, model, use_grad):
    mle, x = unbinned

    log = model is logpdf

    def density(x, n, mu, sigma):
        if log:
            return n, np.log(n) + logpdf(x, mu, sigma)
        return n, n * pdf(x, mu, sigma)

    cost = ExtendedUnbinnedNLL(
        x,
        density,
        verbose=verbose,
        log=log,
        grad=numerical_extended_model_gradient(density),
    )
    assert cost.ndata == np.inf

    m = Minuit(cost, n=len(x), mu=0, sigma=1, grad=use_grad)
    m.limits["n"] = (0, None)
    m.limits["sigma"] = (0, None)
    m.migrad()
    assert m.valid
    assert_allclose(m.values, mle, atol=1e-3)
    assert m.errors["mu"] == pytest.approx(1000**-0.5, rel=0.05)

    assert_equal(m.fmin.reduced_chi2, np.nan)

    if use_grad:
        assert m.ngrad > 0
    else:
        assert m.ngrad == 0


def test_ExtendedUnbinnedNLL_name(unbinned):
    mle, x = unbinned

    cost = ExtendedUnbinnedNLL(
        x, lambda x, *par: (par[0], pdf(x, *par[1:])), name=("n", "mu", "sigma")
    )
    assert cost.ndata == np.inf

    m = Minuit(cost, n=len(x), mu=0, sigma=1)
    assert m.parameters == ("n", "mu", "sigma")
    m.limits["n"] = (0, None)
    m.limits["sigma"] = (0, None)
    m.migrad()
    assert m.valid


@pytest.mark.parametrize("use_grad", (False, True))
def test_ExtendedUnbinnedNLL_2D(use_grad):
    def model(x_y, n, mux, muy, sx, sy, rho):
        return n, n * mvnorm(mux, muy, sx, sy, rho).pdf(x_y.T)

    truth = 100.0, 0.1, 0.2, 0.3, 0.4, 0.5
    x, y = mvnorm(*truth[1:]).rvs(size=int(truth[0]), random_state=1).T

    cost = ExtendedUnbinnedNLL(
        (x, y), model, grad=numerical_extended_model_gradient(model)
    )

    m = Minuit(cost, *truth, grad=use_grad)
    m.limits["n", "sx", "sy"] = (0, None)
    m.limits["rho"] = (-1, 1)
    m.migrad()
    assert m.valid

    assert_allclose(m.values, truth, atol=0.5)

    if use_grad:
        assert m.ngrad > 0

        ref = numerical_cost_gradient(cost)
        assert_allclose(cost.grad(*m.values), ref(*m.values), rtol=0.01)
    else:
        assert m.ngrad == 0


def test_ExtendedUnbinnedNLL_mask():
    def model(x, a):
        return 1, x + a

    c = ExtendedUnbinnedNLL(
        [1, np.nan, 2], model, grad=numerical_extended_model_gradient(model)
    )
    assert c.ndata == np.inf

    assert np.isnan(c(0))
    c.mask = np.arange(3) != 1
    assert not np.isnan(c(0))
    assert c.ndata == np.inf

    ref = numerical_cost_gradient(c)
    assert_allclose(c.grad(0), ref(0))
    assert_allclose(c.grad(1.5), ref(1.5))


@pytest.mark.parametrize("log", (False, True))
def test_ExtendedUnbinnedNLL_properties(log):
    def log_model(x, s, mu, sigma):
        return s, np.log(s) + norm_logpdf(x, mu, sigma)

    def model(x, s, mu, sigma):
        n, y = log_model(x, s, mu, sigma)
        return n, np.exp(y)

    c = ExtendedUnbinnedNLL([1, 2, 3], log_model if log else model, log=log)

    x = np.linspace(0, 1)

    scale, expected = model(x, 1, 2, 3)
    assert_allclose(c.scaled_pdf(x, 1, 2, 3), expected)

    expected /= scale
    assert_allclose(c.pdf(x, 1, 2, 3), expected)

    with pytest.raises(AttributeError):
        c.scaled_pdf = None

    with pytest.raises(AttributeError):
        c.pdf = None


@pytest.mark.parametrize("log", (False, True))
def test_ExtendedUnbinnedNLL_visualize(log):
    pytest.importorskip("matplotlib")

    def log_model(x, s, mu, sigma):
        return s, np.log(s) + norm_logpdf(x, mu, sigma)

    def model(x, s, mu, sigma):
        return s, s * norm_pdf(x, mu, sigma)

    c = ExtendedUnbinnedNLL(
        [1, 2, 3],
        log_model if log else model,
        log=log,
    )

    c.visualize((1, 2, 3))


def test_ExtendedUnbinnedNLL_visualize_2D():
    pytest.importorskip("matplotlib")

    def model(x_y, n, mux, muy, sx, sy, rho):
        return n * 100, n * 100 * mvnorm(mux, muy, sx, sy, rho).pdf(x_y.T)

    truth = 1.0, 0.1, 0.2, 0.3, 0.4, 0.5
    x, y = mvnorm(*truth[1:]).rvs(size=int(truth[0] * 100)).T

    c = ExtendedUnbinnedNLL((x, y), model)

    with pytest.raises(ValueError, match="not implemented for multi-dimensional"):
        c.visualize(truth)


def test_ExtendedUnbinnedNLL_pickle():
    c = ExtendedUnbinnedNLL([1, 2], norm_pdf)
    b = pickle.dumps(c)
    c2 = pickle.loads(b)
    assert_equal(c.data, c2.data)


@pytest.mark.parametrize("verbose", (0, 1))
@pytest.mark.parametrize("use_grad", (False, True))
def test_BinnedNLL(binned, verbose, use_grad):
    mle, nx, xe = binned

    cost = BinnedNLL(nx, xe, cdf, verbose=verbose, grad=numerical_model_gradient(cdf))
    assert cost.ndata == len(nx)

    if use_grad:
        ref = numerical_cost_gradient(cost)
        assert_allclose(cost.grad(*mle[1:]), ref(*mle[1:]))
        assert_allclose(cost.grad(-1, 2), ref(-1, 2))

    m = Minuit(cost, mu=0, sigma=1, grad=use_grad)
    m.limits["sigma"] = (0, None)
    m.migrad()
    assert m.valid
    # binning loses information compared to unbinned case
    assert_allclose(m.values, mle[1:], rtol=0.15)
    assert m.errors["mu"] == pytest.approx(1000**-0.5, rel=0.05)
    assert m.ndof == len(nx) - 2

    assert_allclose(m.fmin.reduced_chi2, 1, atol=0.15)

    if use_grad:
        assert m.ngrad > 0
    else:
        assert m.ngrad == 0


def test_BinnedNLL_pulls(binned):
    mle, nx, xe = binned
    cost = BinnedNLL(nx, xe, cdf)
    m = Minuit(cost, mu=0, sigma=1)
    m.limits["sigma"] = (0, None)
    m.migrad()
    pulls = cost.pulls(m.values)
    assert np.nanmean(pulls) == pytest.approx(0, abs=0.05)
    assert np.nanvar(pulls) == pytest.approx(1, abs=0.2)


@pytest.mark.parametrize("use_grad", (False, True))
def test_BinnedNLL_weighted(use_grad):
    xe = np.array([0, 0.2, 0.4, 0.8, 1.5, 10])
    p = np.diff(expon_cdf(xe, 1))
    n = p * 1000
    c = BinnedNLL(n, xe, expon_cdf)

    assert_equal(c.data, n)
    m1 = Minuit(c, 1)
    m1.migrad()
    assert m1.values[0] == pytest.approx(1, rel=1e-2)

    # variance * 4 = (sigma * 2)^2, constructed so that
    # fitted parameter has twice the uncertainty
    w = np.transpose((n, 4 * n))
    c = BinnedNLL(w, xe, expon_cdf, grad=numerical_model_gradient(expon_cdf))
    assert_equal(c.data, w)

    if use_grad:
        ref = numerical_cost_gradient(c)
        assert_allclose(c.grad(*m1.values), ref(*m1.values))
        assert_allclose(c.grad(12), ref(12))

    m2 = Minuit(c, 1, grad=use_grad)
    m2.migrad()
    assert m2.valid
    assert m2.values[0] == pytest.approx(1, rel=1e-2)
    assert m2.errors[0] == pytest.approx(2 * m1.errors[0], rel=1e-2)

    if use_grad:
        assert m2.ngrad > 0
    else:
        assert m2.ngrad == 0


@pytest.mark.parametrize("use_grad", (False, True))
def test_BinnedNLL_negative_weights(use_grad):
    n = [48.57881659, 7.39393075, -1.6026582, 2.56719152, 1.03007896]
    vn = [242.13921373, 46.13228182, 15.26273872, 6.26334277, 1.52360132]
    xe = [0.0, 1.41453733, 2.82907465, 4.24361198, 5.6581493, 7.07268663]

    w = np.transpose((n, vn))
    c = BinnedNLL(w, xe, expon_cdf, grad=numerical_model_gradient(expon_cdf))

    if use_grad:
        ref = numerical_cost_gradient(c)
        assert_allclose(c.grad(1), ref(1))
        assert_allclose(c.grad(12), ref(12))

    m2 = Minuit(c, 1, grad=use_grad)
    m2.limits = (0, None)
    m2.migrad()
    assert m2.valid
    assert m2.values[0] == pytest.approx(1.3, abs=0.05)

    if use_grad:
        assert m2.ngrad > 0
    else:
        assert m2.ngrad == 0


def test_BinnedNLL_name(binned):
    mle, nx, xe = binned

    cost = BinnedNLL(nx, xe, cdf_nosig, name=("mu", "sigma"))
    assert cost.ndata == len(nx)

    m = Minuit(cost, mu=0, sigma=1)
    m.limits["sigma"] = (0, None)
    m.migrad()
    assert m.valid
    assert m.parameters == ("mu", "sigma")


def test_BinnedNLL_bad_input_1():
    with pytest.raises(ValueError):
        BinnedNLL([1], [1], lambda x, a: 0)


def test_BinnedNLL_bad_input_2():
    with pytest.raises(ValueError):
        BinnedNLL([[[1]]], [1], lambda x, a: 0)


def test_BinnedNLL_bad_input_3():
    with pytest.raises(ValueError):
        BinnedNLL([[1, 2, 3]], [1], lambda x, a: 0)


def test_BinnedNLL_bad_input_4():
    with pytest.raises(ValueError, match="n must either have same dimension as xe"):
        BinnedNLL([[1, 2, 3]], [1, 2], lambda x, a: 0)


def test_BinnedNLL_bad_input_5():
    with pytest.raises(ValueError, match="n must either have same dimension as xe"):
        BinnedNLL([[1, 2, 3]], [1, 2], lambda x, a: 0)


def test_BinnedNLL_bad_input_6():
    with pytest.raises(
        ValueError, match="not supported for multidimensional histograms"
    ):
        BinnedNLL([[1]], [[1, 2], [3, 4]], lambda x, a: 0, use_pdf="numerical")


def test_BinnedNLL_bad_input_7():
    with pytest.raises(ValueError, match="xe must be iterable"):
        BinnedNLL([0, 1], 0, lambda x, a: 0)


def test_BinnedNLL_bad_input_8():
    with pytest.raises(ValueError, match="use_pdf and grad cannot be used together"):
        BinnedNLL(
            [0, 1], [1, 2, 3], lambda x, a: 0, grad=lambda x, a: [1], use_pdf="numeric"
        )


def test_BinnedNLL_ndof_zero():
    c = BinnedNLL([1], [0, 1], lambda x, scale: expon_cdf(x, scale))
    m = Minuit(c, scale=1)
    m.migrad()
    assert c.ndata == m.nfit
    assert np.isnan(m.fmin.reduced_chi2)


@pytest.mark.parametrize("use_grad", (False, True))
def test_BinnedNLL_2D(use_grad):
    truth = (0.1, 0.2, 0.3, 0.4, 0.5)
    x, y = mvnorm(*truth).rvs(size=1000, random_state=1).T

    w, xe, ye = np.histogram2d(x, y, bins=(20, 50))

    def model(xy, mux, muy, sx, sy, rho):
        return mvnorm(mux, muy, sx, sy, rho).cdf(xy.T)

    cost = BinnedNLL(w, (xe, ye), model, grad=numerical_model_gradient(model))
    assert cost.ndata == np.prod(w.shape)

    if use_grad:
        ref = numerical_cost_gradient(cost)
        assert_allclose(cost.grad(*truth), ref(*truth))

    m = Minuit(cost, *truth, grad=use_grad)
    m.limits["sx", "sy"] = (0, None)
    m.limits["rho"] = (-1, 1)
    m.migrad()
    assert m.valid
    assert_allclose(m.values, truth, atol=0.05)

    if use_grad:
        assert m.ngrad > 0
    else:
        assert m.ngrad == 0

    assert cost.ndata == np.prod(w.shape)
    w2 = w.copy()
    w2[1, 1] += 1
    cost.n = w2
    assert cost(*m.values) > m.fval


def test_BinnedNLL_2D_with_zero_bins():
    truth = (0.1, 0.2, 0.3, 0.4, 0.5)
    x, y = mvnorm(*truth).rvs(size=1000, random_state=1).T

    w, xe, ye = np.histogram2d(x, y, bins=(50, 100), range=((-5, 5), (-5, 5)))
    assert np.mean(w == 0) > 0.25

    def model(xy, mux, muy, sx, sy, rho):
        return mvnorm(mux, muy, sx, sy, rho).cdf(xy.T)

    cost = BinnedNLL(w, (xe, ye), model)
    m = Minuit(cost, *truth)
    m.limits["sx", "sy"] = (0, None)
    m.limits["rho"] = (-1, 1)
    m.migrad()
    assert m.valid
    assert_allclose(m.values, truth, atol=0.05)


def test_BinnedNLL_mask():
    c = BinnedNLL(
        [5, 1000, 1], [0, 1, 2, 3], expon_cdf, grad=numerical_model_gradient(expon_cdf)
    )
    assert c.ndata == 3

    ref = numerical_cost_gradient(c)
    assert_allclose(c.grad(2), ref(2))

    c_unmasked = c(1)
    c.mask = np.arange(3) != 1
    assert c(1) < c_unmasked
    assert c.ndata == 2

    ref = numerical_cost_gradient(c)
    assert_allclose(c.grad(2), ref(2))


def test_BinnedNLL_properties():
    def cdf(x, a, b):
        return 0

    c = BinnedNLL([1], [1, 2], cdf)
    assert c.cdf is cdf
    with pytest.raises(AttributeError):
        c.cdf = None
    assert_equal(c.n, [1])
    assert_equal(c.xe, [1, 2])
    c.n = [2]
    assert_equal(c.n, [2])
    with pytest.raises(ValueError):
        c.n = [1, 2]


def test_BinnedNLL_visualize():
    pytest.importorskip("matplotlib")

    c = BinnedNLL([1, 2], [1, 2, 3], expon_cdf)

    c.visualize((1,))

    c.mask = np.array([False, True])
    c.visualize((1,))


def test_BinnedNLL_visualize_2D():
    pytest.importorskip("matplotlib")

    truth = (0.1, 0.2, 0.3, 0.4, 0.5)
    x, y = mvnorm(*truth).rvs(size=10, random_state=1).T
    w, xe, ye = np.histogram2d(x, y, bins=(50, 100), range=((-5, 5), (-5, 5)))

    def model(xy, mux, muy, sx, sy, rho):
        return mvnorm(mux, muy, sx, sy, rho).cdf(xy.T)

    c = BinnedNLL(w, (xe, ye), model)

    c.visualize(truth)


def test_BinnedNLL_pickle():
    c = BinnedNLL([1], [1, 2], expon_cdf)
    b = pickle.dumps(c)
    c2 = pickle.loads(b)
    assert_equal(c.data, c2.data)


@pytest.mark.parametrize("use_pdf", ["approximate", "numerical"])
def test_BinnedNLL_with_pdf(use_pdf):
    if use_pdf == "numerical":
        pytest.importorskip("scipy")

    xe = np.array([0, 0.1, 0.2, 0.3])
    n = [1, 2, 3]
    c = BinnedNLL(n, xe, norm_pdf, use_pdf=use_pdf)

    par = [0.05, 1]
    ref = np.diff(norm_cdf(xe, *par)) * np.sum(n)
    assert_allclose(c.prediction(par), ref, rtol=1e-3)


def test_BinnedNLL_use_pdf_bad_value():
    with pytest.raises(ValueError):
        BinnedNLL([1, 2], [1, 2, 3], norm_pdf, use_pdf="foo")


@pytest.mark.parametrize("use_pdf", ["approximate", "numerical"])
def test_BinnedNLL_with_pdf_3D(use_pdf):
    xe = (np.array([0, 0.1]), np.array([0.1, 0.3, 0.4]), np.array([0, 0.1, 0.2, 0.3]))
    n = [[[1, 2, 3], [3, 4, 5]]]

    def pdf(r, mx, my, mz, sx, sy, sz):
        x, y, z = r
        return norm_pdf(x, mx, sx) * norm_pdf(y, my, sy) * norm_pdf(z, mz, sz)

    def cdf(r, mx, my, mz, sx, sy, sz):
        x, y, z = r
        return norm_cdf(x, mx, sx) * norm_cdf(y, my, sy) * norm_cdf(z, mz, sz)

    if use_pdf == "numerical":
        with pytest.raises(
            ValueError,
            match=(
                'use_pdf="numerical" is not supported '
                "for multidimensional histograms"
            ),
        ):
            BinnedNLL(n, xe, pdf, use_pdf=use_pdf)
    else:
        c1 = BinnedNLL(n, xe, pdf, use_pdf=use_pdf)
        c2 = BinnedNLL(n, xe, cdf)

        par = [0, 1, 2, 3, 4, 5]
        assert_allclose(c1.prediction(par), c2.prediction(par), rtol=1e-3)


@pytest.mark.parametrize("verbose", (0, 1))
@pytest.mark.parametrize("use_grad", (False, True))
def test_ExtendedBinnedNLL(binned, verbose, use_grad):
    mle, nx, xe = binned

    cost = ExtendedBinnedNLL(
        nx, xe, scaled_cdf, verbose=verbose, grad=numerical_model_gradient(scaled_cdf)
    )
    assert cost.ndata == len(nx)

    if use_grad:
        ref = numerical_cost_gradient(cost)
        assert_allclose(cost.grad(*mle), ref(*mle))

    m = Minuit(cost, n=mle[0], mu=0, sigma=1, grad=use_grad)
    m.limits["n"] = (0, None)
    m.limits["sigma"] = (0, None)
    m.migrad()
    assert m.valid
    # binning loses information compared to unbinned case
    assert_allclose(m.values, mle, rtol=0.15)
    assert m.errors["mu"] == pytest.approx(1000**-0.5, rel=0.05)
    assert m.ndof == len(nx) - 3

    assert_allclose(m.fmin.reduced_chi2, 1, 0.1)

    if use_grad:
        assert m.ngrad > 0
    else:
        assert m.ngrad == 0


def test_ExtendedBinnedNLL_name(binned):
    mle, nx, xe = binned

    cost = ExtendedBinnedNLL(nx, xe, scaled_cdf_nosig, name=("n", "mu", "sigma"))
    assert cost.ndata == len(nx)

    m = Minuit(cost, n=mle[0], mu=0, sigma=1)
    m.limits["sigma"] = (0, None)
    m.migrad()
    assert m.valid
    assert m.parameters == ("n", "mu", "sigma")


@pytest.mark.parametrize("use_grad", (False, True))
def test_ExtendedBinnedNLL_weighted(use_grad):
    xe = np.array([0, 1, 10])
    n = np.diff(expon_cdf(xe, 1))
    m1 = Minuit(ExtendedBinnedNLL(n, xe, expon_cdf), 1)
    m1.migrad()
    assert_allclose(m1.values, (1,), rtol=1e-2)

    w = np.transpose((n, 4 * n))
    c = ExtendedBinnedNLL(w, xe, expon_cdf, grad=numerical_model_gradient(expon_cdf))
    if use_grad:
        ref = numerical_cost_gradient(c)
        assert_allclose(c.grad(1), ref(1), atol=1e-14)
        assert_allclose(c.grad(12), ref(12))
    m2 = Minuit(c, 1.1, grad=use_grad)
    m2.migrad()
    assert m2.valid
    assert_allclose(m2.values, (1,), rtol=1e-2)

    assert m2.errors[0] == pytest.approx(2 * m1.errors[0], rel=1e-2)

    if use_grad:
        assert m2.ngrad > 0
    else:
        assert m2.ngrad == 0


@pytest.mark.parametrize("use_grad", (False, True))
def test_ExtendedBinnedNLL_negative_weights(use_grad):
    n = [48.57881659, 7.39393075, -1.6026582, 2.56719152, 1.03007896]
    vn = [242.13921373, 46.13228182, 15.26273872, 6.26334277, 1.52360132]
    xe = [0.0, 1.41453733, 2.82907465, 4.24361198, 5.6581493, 7.07268663]

    w = np.transpose((n, vn))
    c = ExtendedBinnedNLL(
        w, xe, scaled_expon_cdf, grad=numerical_model_gradient(scaled_expon_cdf)
    )

    # if use_grad:
    #     ref = numerical_cost_gradient(c)
    #     assert_allclose(c.grad(1, 0.1), ref(1, 0.1))
    #     assert_allclose(c.grad(1, 1), ref(1, 1))
    #     assert_allclose(c.grad(2, 12), ref(2, 12))

    m2 = Minuit(c, 50, 1, grad=use_grad)
    m2.limits = (0, None)
    m2.migrad()
    assert m2.valid
    assert m2.values[0] == pytest.approx(65, abs=1)
    assert m2.values[1] == pytest.approx(1.3, abs=0.05)

    if use_grad:
        assert m2.ngrad > 0
    else:
        assert m2.ngrad == 0


def test_ExtendedBinnedNLL_bad_input():
    with pytest.raises(ValueError):
        ExtendedBinnedNLL([1], [1], lambda x, a: 0)


@pytest.mark.parametrize("use_grad", (False, True))
def test_ExtendedBinnedNLL_2D(use_grad):
    truth = (1.0, 0.1, 0.2, 0.3, 0.4, 0.5)
    x, y = mvnorm(*truth[1:]).rvs(size=int(truth[0] * 100), random_state=1).T

    w, xe, ye = np.histogram2d(x, y, bins=(10, 20))

    def model(xy, n, mux, muy, sx, sy, rho):
        return n * 100 * mvnorm(mux, muy, sx, sy, rho).cdf(np.transpose(xy))

    cost = ExtendedBinnedNLL(w, (xe, ye), model, grad=numerical_model_gradient(model))
    assert cost.ndata == np.prod(w.shape)

    if use_grad:
        ref = numerical_cost_gradient(cost)
        assert_allclose(cost.grad(*truth), ref(*truth))

    m = Minuit(cost, *truth, grad=use_grad)
    m.limits["n", "sx", "sy"] = (0, None)
    m.limits["rho"] = (-1, 1)
    m.migrad()
    assert m.valid
    assert_allclose(m.values, truth, atol=0.1)

    if use_grad:
        assert m.ngrad > 0
    else:
        assert m.ngrad == 0


def test_ExtendedBinnedNLL_3D():
    norm = pytest.importorskip("scipy.stats").norm

    truth = (1.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7)
    n = int(truth[0] * 10000)
    x, y = mvnorm(*truth[1:-2]).rvs(size=n).T
    z = norm(truth[-2], truth[-1]).rvs(size=n)

    w, edges = np.histogramdd((x, y, z), bins=(5, 10, 20))

    def model(xyz, n, mux, muy, sx, sy, rho, muz, sz):
        *xy, z = xyz
        return (
            n
            * 10000
            * mvnorm(mux, muy, sx, sy, rho).cdf(np.transpose(xy))
            * norm(muz, sz).cdf(z)
        )

    cost = ExtendedBinnedNLL(w, edges, model)
    assert cost.ndata == np.prod(w.shape)
    m = Minuit(cost, *truth)
    m.limits["n", "sx", "sy", "sz"] = (0, None)
    m.limits["rho"] = (-1, 1)
    m.migrad()
    assert m.valid
    assert_allclose(m.values, truth, atol=0.05)


def test_ExtendedBinnedNLL_mask():
    c = ExtendedBinnedNLL(
        [1, 1000, 2], [0, 1, 2, 3], expon_cdf, grad=numerical_model_gradient(expon_cdf)
    )
    assert c.ndata == 3

    c_unmasked = c(2)
    c.mask = np.arange(3) != 1
    assert c(2) < c_unmasked
    assert c.ndata == 2

    ref = numerical_cost_gradient(c)
    assert_allclose(c.grad(2), ref(2))


def test_ExtendedBinnedNLL_properties():
    def cdf(x, a):
        return 0

    c = ExtendedBinnedNLL([1], [1, 2], cdf)
    assert c.scaled_cdf is cdf


def test_ExtendedBinnedNLL_visualize():
    pytest.importorskip("matplotlib")

    def model(x, s, slope):
        return s * expon_cdf(x, slope)

    c = ExtendedBinnedNLL([1], [1, 2], model)
    c.visualize((1, 2))


def test_ExtendedBinnedNLL_visualize_2D():
    pytest.importorskip("matplotlib")

    truth = (1.0, 0.1, 0.2, 0.3, 0.4, 0.5)
    x, y = mvnorm(*truth[1:]).rvs(size=int(truth[0] * 1000), random_state=1).T

    w, xe, ye = np.histogram2d(x, y, bins=(10, 20))

    def model(xy, n, mux, muy, sx, sy, rho):
        return n * 1000 * mvnorm(mux, muy, sx, sy, rho).cdf(np.transpose(xy))

    c = ExtendedBinnedNLL(w, (xe, ye), model)

    c.visualize(truth)


def test_ExtendedBinnedNLL_pickle():
    c = BinnedNLL([1], [1, 2], expon_cdf)
    b = pickle.dumps(c)
    c2 = pickle.loads(b)
    assert_equal(c.data, c2.data)


def test_ExtendedBinnedNLL_weighted_pulls():
    rng = np.random.default_rng(1)
    xe = np.linspace(0, 5, 500)
    p = np.diff(expon_cdf(xe, 2))
    m = p * 100000
    n = rng.poisson(m * 4) / 4
    nvar = m * 4 / 4**2
    w = np.transpose((n, nvar))
    c = ExtendedBinnedNLL(w, xe, lambda x, n, s: n * expon_cdf(x, s))
    m = Minuit(c, np.sum(n), 2)
    m.limits["n"] = (0.1, None)
    m.limits["s"] = (0.1, None)
    m.migrad()
    assert m.valid
    pulls = c.pulls(m.values)
    assert np.nanmean(pulls) == pytest.approx(0, abs=0.1)
    assert np.nanvar(pulls) == pytest.approx(1, abs=0.2)


@pytest.mark.parametrize(
    "loss,use_grad",
    [
        ("linear", False),
        ("linear", True),
        ("soft_l1", False),
        ("soft_l1", True),
        (np.arctan, False),
    ],
)
@pytest.mark.parametrize("verbose", (0, 1))
def test_LeastSquares(loss, verbose, use_grad):
    rng = np.random.default_rng(1)

    x = np.linspace(0, 1, 1000)
    ye = 0.1
    y = rng.normal(2 * x + 1, ye)

    cost = LeastSquares(
        x,
        y,
        ye,
        line,
        loss=loss,
        verbose=verbose,
        grad=numerical_model_gradient(line),
    )
    assert cost.ndata == len(x)

    if use_grad:
        ref = numerical_cost_gradient(cost)
        assert_allclose(cost.grad(1, 2), ref(1, 2))

    m = Minuit(cost, a=0, b=0, grad=use_grad)
    m.migrad()
    assert_allclose(m.values, (1, 2), rtol=0.05)
    assert cost.loss == loss
    if loss != "linear":
        cost.loss = "linear"
        assert cost.loss != loss
    m.migrad()
    assert_allclose(m.values, (1, 2), rtol=0.05)
    assert m.ndof == len(x) - 2

    assert_allclose(m.fmin.reduced_chi2, 1, atol=5e-2)

    if use_grad:
        assert m.ngrad > 0
    else:
        assert m.ngrad == 0


def test_LeastSquares_name():
    rng = np.random.default_rng(1)

    x = np.linspace(0, 1, 1000)
    ye = 0.1
    y = rng.normal(2 * x + 1, ye)

    cost = LeastSquares(x, y, ye, line_nosig, name=("a", "b"))
    assert cost.ndata == len(x)

    m = Minuit(cost, a=0, b=0)
    assert m.parameters == ("a", "b")
    m.migrad()
    assert_allclose(m.values, (1, 2), rtol=0.05)


def test_LeastSquares_2D():
    def model(xy, a, b):
        x, y = xy
        return a * x + b * y

    x = np.array([1.0, 2.0, 3.0])
    y = np.array([4.0, 5.0, 6.0])
    f = model((x, y), 1.5, 0.2)
    fe = 1.5

    c = LeastSquares((x, y), f, fe, model, grad=numerical_model_gradient(model))
    assert c.ndata == 3

    ref = numerical_cost_gradient(c)
    assert_allclose(c.grad(1, 2), ref(1, 2))

    assert_equal(c.x, (x, y))
    assert_equal(c.y, f)
    assert_equal(c.yerror, fe)
    assert_allclose(c(1.5, 0.2), 0.0)
    assert_allclose(c(2.5, 0.2), np.sum(((f - 2.5 * x - 0.2 * y) / fe) ** 2))
    assert_allclose(c(1.5, 1.2), np.sum(((f - 1.5 * x - 1.2 * y) / fe) ** 2))

    c.y = 2 * f
    assert_equal(c.y, 2 * f)
    c.x = (y, x)
    assert_equal(c.x, (y, x))


def test_LeastSquares_3D():
    def model(xyz, a, b):
        x, y, z = xyz
        return a * x + b * y + a * b * z

    x = np.array([1.0, 2.0, 3.0, 4.0])
    y = np.array([4.0, 5.0, 6.0, 7.0])
    z = np.array([7.0, 8.0, 9.0, 10.0])

    f = model((x, y, z), 1.5, 0.2)
    fe = 1.5

    c = LeastSquares((x, y, z), f, fe, model, grad=numerical_model_gradient(model))
    assert c.ndata == 4

    ref = numerical_cost_gradient(c)
    assert_allclose(c.grad(1, 2), ref(1, 2))

    assert_equal(c.x, (x, y, z))
    assert_equal(c.y, f)
    assert_equal(c.yerror, fe)
    assert_allclose(c(1.5, 0.2), 0.0)
    assert_allclose(
        c(2.5, 0.2), np.sum(((f - 2.5 * x - 0.2 * y - 2.5 * 0.2 * z) / fe) ** 2)
    )
    assert_allclose(
        c(1.5, 1.2), np.sum(((f - 1.5 * x - 1.2 * y - 1.5 * 1.2 * z) / fe) ** 2)
    )

    c.y = 2 * z
    assert_equal(c.y, 2 * z)
    c.x = (y, x, z)
    assert_equal(c.x, (y, x, z))


def test_LeastSquares_bad_input():
    with pytest.raises(ValueError, match="shape mismatch"):
        LeastSquares([1, 2], [], [1], lambda x, a: 0)

    with pytest.raises(ValueError, match="shape mismatch"):
        LeastSquares([1, 2], [3, 4, 5], [1], lambda x, a: 0)

    with pytest.raises(ValueError, match="unknown loss"):
        LeastSquares([1], [1], [1], lambda x, a: 0, loss="foo")

    with pytest.raises(ValueError, match="loss must be str or LossFunction"):
        LeastSquares([1], [1], [1], lambda x, a: 0, loss=[1, 2, 3])


@pytest.mark.parametrize("use_grad", (False, True))
def test_LeastSquares_mask(use_grad):
    c = LeastSquares(
        [1, 2, 3],
        [3, np.nan, 4],
        [1, 1, 1],
        lambda x, a: x + a,
        grad=lambda x, a: np.ones_like(x),
    )
    assert c.ndata == 3
    assert np.isnan(c(0))
    assert np.all(np.isnan(c.grad(1)))

    m = Minuit(c, 1, grad=use_grad)
    assert m.ndof == 2
    m.migrad()
    assert not m.valid

    c.mask = np.arange(3) != 1
    assert c.ndata == 2
    assert not np.isnan(c(0))
    assert not np.any(np.isnan(c.grad(0)))

    assert m.ndof == 1
    m.migrad()
    assert m.valid
    assert_equal(m.values, [1.5])


def test_LeastSquares_mask_2():
    c = LeastSquares([1, 2], [1, 5], 1, lambda x, a: a * x)
    assert c(2) == pytest.approx(2)
    c.mask = [False, True]
    assert c(2) == pytest.approx(1)
    c.x = [2, 2]
    c.y = [4, 4]
    c.yerror = [2, 2]
    assert c(2) == pytest.approx(0)
    assert c(1) == pytest.approx(1)


def test_LeastSquares_properties():
    def model(x, a):
        return x + 2 * a

    c = LeastSquares(1, 2, 3, model)
    assert_equal(c.x, [1])
    assert_equal(c.y, [2])
    assert_equal(c.yerror, [3])
    assert c.model(1, 1) == model(1, 1)
    assert c.model(2, 3) == model(2, 3)
    with pytest.raises(AttributeError):
        c.model = model
    with pytest.raises(ValueError):
        c.x = [1, 2]
    with pytest.raises(ValueError):
        c.y = [1, 2]
    with pytest.raises(ValueError):
        c.yerror = [1, 2]


def test_LeastSquares_visualize():
    pytest.importorskip("matplotlib")

    c = LeastSquares([1, 2], [2, 3], 0.1, line)

    # auto-sampling
    (x, y, ye), (xm, ym) = c.visualize((1, 2))
    assert_equal(x, (1, 2))
    assert_equal(y, (2, 3))
    assert_equal(ye, 0.1)
    assert len(xm) < 10
    # linear spacing
    (x, y, ye), (xm, ym) = c.visualize((1, 2), model_points=10)
    assert len(xm) == 10
    assert_allclose(xm[1:] - xm[:-1], xm[1] - xm[0])
    # trigger use of log-spacing
    c = LeastSquares([1, 10, 100], [2, 3, 4], 0.1, line)
    (x, y, ye), (xm, ym) = c.visualize((1, 2), model_points=10)
    assert len(xm) == 10
    assert_allclose(xm[1:] / xm[:-1], xm[1] / xm[0])
    # manual spacing
    (x, y, ye), (xm, ym) = c.visualize((1, 2), model_points=np.linspace(1, 100))
    assert_equal(xm, np.linspace(1, 100))


def test_LeastSquares_visualize_par_array():
    pytest.importorskip("matplotlib")

    def line(x, par):
        return par[0] + par[1] * x

    c = LeastSquares([1, 2], [2, 3], 0.1, line)

    c.visualize((1, 2))


def test_LeastSquares_visualize_2D():
    pytest.importorskip("matplotlib")

    c = LeastSquares([[1, 2], [2, 3]], [1, 2], 0.1, line)

    with pytest.raises(ValueError, match="not implemented for multi-dimensional"):
        c.visualize((1, 2))


def test_LeastSquares_pickle():
    c = LeastSquares([1, 2], [2, 3], 0.1, line)

    b = pickle.dumps(c)
    c2 = pickle.loads(b)
    assert_equal(c.data, c2.data)
    assert c.model is c2.model


def test_LeastSquares_pulls():
    c = LeastSquares([1, 2], [2, 3], 0.1, line)
    # pulls are zero for parameters 1, 1
    assert_equal(c.pulls((1, 1)), [0, 0])
    # correct slope but line offset by 1 with error 0.1 produces pulls [10, 10]
    assert_equal(c.pulls((0, 1)), [10, 10])
    c.mask = [True, False]
    assert_equal(c.pulls((0, 1)), [10, np.nan])


@pytest.mark.parametrize("use_grad", (False, True))
def test_CostSum_1(use_grad):
    def model1(x, a):
        return a + x

    def grad1(x, a):
        return np.ones((1, len(x)))

    def model2(x, b, a):
        return a + b * x

    def grad2(x, b, a):
        g = np.empty((2, len(x)))
        g[0] = x
        g[1] = 1
        return g

    def model3(x, c):
        return c * np.ones_like(x)

    def grad3(x, c):
        return np.ones((1, len(x)))

    lsq1 = LeastSquares(1, 2, 3, model1, grad=grad1)
    assert describe(lsq1) == ["a"]

    if use_grad:
        assert_allclose(lsq1.grad(2), numerical_cost_gradient(lsq1)(2))

    lsq2 = LeastSquares(1, 3, 4, model2, grad=grad2)
    assert describe(lsq2) == ["b", "a"]

    if use_grad:
        assert_allclose(lsq2.grad(2, 3), numerical_cost_gradient(lsq2)(2, 3))

    lsq3 = LeastSquares(1, 1, 1, model3, grad=grad3)
    assert describe(lsq3) == ["c"]

    if use_grad:
        assert_allclose(lsq3.grad(4), numerical_cost_gradient(lsq3)(4))

    lsq12 = lsq1 + lsq2
    assert lsq12._items == [lsq1, lsq2]
    assert isinstance(lsq12, CostSum)
    assert isinstance(lsq1, LeastSquares)
    assert isinstance(lsq2, LeastSquares)
    assert describe(lsq12) == ["a", "b"]
    assert lsq12.ndata == 2

    assert lsq12(1, 2) == lsq1(1) + lsq2(2, 1)

    if use_grad:
        a = 2
        b = 3
        ref = np.zeros(2)
        ref[0] += lsq1.grad(a)[0]
        ref[[1, 0]] += lsq2.grad(b, a)
        assert_allclose(lsq12.grad(a, b), ref)

    lsq121 = lsq12 + lsq1
    assert lsq121._items == [lsq1, lsq2, lsq1]
    assert describe(lsq121) == ["a", "b"]
    assert lsq121.ndata == 3

    if use_grad:
        a = 2
        b = 3
        ref = np.zeros(2)
        ref[0] += lsq1.grad(a)[0]
        ref[[1, 0]] += lsq2.grad(b, a)
        ref[0] += lsq1.grad(a)[0]
        assert_allclose(lsq121.grad(a, b), ref)

    lsq312 = lsq3 + lsq12
    assert lsq312._items == [lsq3, lsq1, lsq2]
    assert describe(lsq312) == ["c", "a", "b"]
    assert lsq312.ndata == 3

    if use_grad:
        a = 2
        b = 3
        c = 4
        ref = np.zeros(3)
        ref[0] += lsq3.grad(c)[0]
        ref[1] += lsq1.grad(a)[0]
        ref[[2, 1]] += lsq2.grad(b, a)
        assert_allclose(lsq312.grad(c, a, b), ref)

    lsq31212 = lsq312 + lsq12
    assert lsq31212._items == [lsq3, lsq1, lsq2, lsq1, lsq2]
    assert describe(lsq31212) == ["c", "a", "b"]
    assert lsq31212.ndata == 5

    if use_grad:
        a = 2
        b = 3
        c = 4
        ref = np.zeros(3)
        ref[:] += lsq312.grad(c, a, b)
        ref[1:] += lsq12.grad(a, b)
        assert_allclose(lsq31212.grad(c, a, b), ref)

    lsq31212 += lsq1
    assert lsq31212._items == [lsq3, lsq1, lsq2, lsq1, lsq2, lsq1]
    assert describe(lsq31212) == ["c", "a", "b"]
    assert lsq31212.ndata == 6

    m = Minuit(lsq12, a=0, b=0, grad=use_grad)
    m.migrad()
    assert m.parameters == ("a", "b")
    assert_allclose(m.values, (1, 2))
    assert_allclose(m.errors, (3, 5))
    assert_allclose(m.covariance, ((9, -9), (-9, 25)), atol=1e-10)

    if use_grad:
        assert m.ngrad > 0
    else:
        assert m.ngrad == 0


def test_CostSum_2():
    ref = NormalConstraint("a", 1, 2), NormalConstraint(("b", "a"), (1, 1), (2, 2))
    cs = ref[0] + ref[1]
    assert cs.ndata == 3
    assert isinstance(cs, Sequence)
    assert len(cs) == 2
    assert cs[0] is ref[0]
    assert cs[1] is ref[1]
    for c, r in zip(cs, ref):
        assert c is r
    assert cs.index(ref[0]) == 0
    assert cs.index(ref[1]) == 1
    assert cs.count(ref[0]) == 1


def test_CostSum_3():
    def line_np(x, par):
        return par[0] + par[1] * x

    lsq = LeastSquares([1, 2, 3], [3, 4, 5], 1, line_np)
    con = NormalConstraint("par", (1, 1), (1, 1))
    cs = sum([lsq, con])
    assert cs((1, 1)) == lsq((1, 1)) + con((1, 1))

    cs = 1.5 + lsq + con
    assert cs((1, 1)) == lsq((1, 1)) + con((1, 1)) + 1.5


def test_CostSum_4():
    pytest.importorskip("scipy.special")

    t = Template([1, 2], [1, 2, 3], [[1, 1], [0, 1]], method="asy")
    assert t.errordef == Minuit.LIKELIHOOD

    m1 = Minuit(t, 1, 1)
    m1.migrad()

    cs = CostSum(t)
    assert cs.errordef == Minuit.LEAST_SQUARES

    m2 = Minuit(cs, 1, 1)
    m2.migrad()

    assert_allclose(m1.errors, m2.errors)


def test_CostSum_visualize():
    pytest.importorskip("matplotlib")

    lsq = LeastSquares([1, 2, 3], [3, 4, 5], 1, line)
    con = NormalConstraint(("a", "b"), (1, 1), (1, 1))
    c = lsq + con + 1
    c.visualize((1, 2))


def test_NormalConstraint_1():
    c1 = NormalConstraint("a", 1, 1.5)
    c2 = NormalConstraint(("a", "b"), (1, 2), (3, 4))
    c3 = NormalConstraint(("a", "b"), (1, 2), ((1, 0.5), (0.5, 4)))

    assert describe(c1) == ["a"]
    assert describe(c2) == ["a", "b"]
    assert describe(c3) == ["a", "b"]
    assert c1.ndata == 1
    assert c2.ndata == 2
    assert c3.ndata == 2
    assert c1.has_grad
    assert c2.has_grad
    assert c3.has_grad
    assert c1.value == 1
    assert_equal(c1.covariance, [1.5**2])
    assert_equal(c2.value, (1, 2))
    assert_equal(c2.covariance, (9, 16))
    assert_equal(c3.value, (1, 2))
    assert_equal(c3.covariance, ((1, 0.5), (0.5, 4)))

    assert_allclose(c1(1), 0)
    assert_allclose(c1(1 + 1.5), 1)
    assert_allclose(c1(1 - 1.5), 1)

    assert_allclose(c2(1, 2), 0)
    assert_allclose(c2(1 + 3, 2), 1)
    assert_allclose(c2(1 - 3, 2), 1)
    assert_allclose(c2(1, 2 + 4), 1)
    assert_allclose(c2(1, 2 - 4), 1)

    def ref(x, y):
        d = np.subtract((x, y), (1, 2))
        c = ((1, 0.5), (0.5, 4))
        cinv = np.linalg.inv(c)
        return d.T @ cinv @ d

    assert_allclose(c3(1, 2), 0)
    assert_allclose(c3(2, 2), ref(2, 2))
    assert_allclose(c3(3, 4), ref(3, 4))
    assert_allclose(c3(-1, -2), ref(-1, -2))

    ref = numerical_cost_gradient(c1)
    assert_allclose(c1.grad(1), [0])
    assert_allclose(c1.grad(2), ref(2))
    assert_allclose(c1.grad(-2), ref(-2))

    ref = numerical_cost_gradient(c2)
    assert_allclose(c2.grad(1, 2), [0, 0])
    assert_allclose(c2.grad(2, 3), ref(2, 3))
    assert_allclose(c2.grad(-2, -4), ref(-2, -4))

    ref = numerical_cost_gradient(c3)
    assert_allclose(c3.grad(1, 2), [0, 0])
    assert_allclose(c3.grad(2, 3), ref(2, 3))
    assert_allclose(c3.grad(-2, -4), ref(-2, -4))

    c1.value = 2
    c1.covariance = 4
    assert_equal(c1.value, 2)
    assert_equal(c1.covariance, 4)
    assert_allclose(c1(2), 0)
    assert_allclose(c1(4), 1)
    assert_allclose(c1(0), 1)

    c2.value = (2, 3)
    c2.covariance = (1, 4)
    assert_equal(c2.value, (2, 3))
    assert_equal(c2.covariance, (1, 4))
    assert_allclose(c2(2, 3), 0)
    assert_allclose(c2(1, 3), 1)
    assert_allclose(c2(2, 5), 1)

    c3.value = (2, 3)
    c3.covariance = [(4, 1), (1, 5)]
    assert_equal(c3.value, (2, 3))
    assert_equal(c3.covariance, [(4, 1), (1, 5)])

    def ref(x, y):
        d = np.subtract((x, y), (2, 3))
        c = ((4, 1), (1, 5))
        cinv = np.linalg.inv(c)
        return d.T @ cinv @ d

    assert_allclose(c3(2, 3), 0)
    assert_allclose(c3(1, 3), ref(1, 3))
    assert_allclose(c3(2, 5), ref(2, 5))


@pytest.mark.parametrize("use_grad", (False, True))
def test_NormalConstraint_2(use_grad):
    c1 = NormalConstraint("a", 1, 1)
    c2 = c1 + NormalConstraint("a", 1, 0.1)

    m = Minuit(c1, 0, grad=use_grad)
    m.migrad()
    if use_grad:
        assert m.ngrad > 0
    assert_allclose(m.values, (1,), atol=1e-2)
    assert_allclose(m.errors, (1,), rtol=1e-2)

    m = Minuit(c2, 0, grad=use_grad)
    m.migrad()
    if use_grad:
        assert m.ngrad > 0
    assert_allclose(m.values, (1,), atol=1e-2)
    assert_allclose(m.errors, (0.1,), rtol=1e-2)


def test_NormalConstraint_3():
    lsq1 = NormalConstraint(("a", "b"), (1, 2), (2, 2))
    lsq2 = lsq1 + NormalConstraint("b", 2, 0.1) + NormalConstraint("a", 1, 0.01)
    sa = 0.1
    sb = 0.02
    rho = 0.5
    cov = ((sa**2, rho * sa * sb), (rho * sa * sb, sb**2))
    lsq3 = lsq1 + NormalConstraint(("a", "b"), (1, 2), cov)
    assert describe(lsq1) == ["a", "b"]
    assert describe(lsq2) == ["a", "b"]
    assert describe(lsq3) == ["a", "b"]
    assert lsq1.ndata == 2
    assert lsq2.ndata == 4

    m = Minuit(lsq1, 0, 0)
    m.migrad()
    assert_allclose(m.values, (1, 2), atol=1e-3)
    assert_allclose(m.errors, (2, 2), rtol=1e-3)

    m = Minuit(lsq2, 0, 0)
    m.migrad()
    assert_allclose(m.values, (1, 2), atol=1e-3)
    assert_allclose(m.errors, (0.01, 0.1), rtol=1e-2)

    m = Minuit(lsq3, 0, 0)
    m.migrad()
    assert_allclose(m.values, (1, 2), atol=1e-3)
    assert_allclose(m.errors, (sa, sb), rtol=1e-2)
    assert_allclose(m.covariance, cov, rtol=1e-2)


def test_NormalConstraint_visualize():
    pytest.importorskip("matplotlib")

    c = NormalConstraint(("a", "b"), (1, 2), (3, 4))
    c.visualize((1, 2))

    c = NormalConstraint(("a", "b"), (1, 2), np.eye(2))
    c.visualize((1, 2))


def test_NormalConstraint_pickle():
    c = NormalConstraint(("a", "b"), (1, 2), (3, 4))
    b = pickle.dumps(c)
    c2 = pickle.loads(b)
    assert describe(c) == describe(c2) and describe(c) == ["a", "b"]
    assert_equal(c.value, c2.value)
    assert_equal(c.covariance, c2.covariance)


def test_NormalConstraint_bad_input_1():
    with pytest.raises(ValueError, match="scalar or one-dimensional"):
        NormalConstraint("par", [[[1, 2]]], np.eye(2))


def test_NormalConstraint_bad_input_2():
    with pytest.raises(ValueError, match="not match size of args"):
        NormalConstraint(["a", "b", "c"], [1, 2], np.eye(2))


def test_NormalConstraint_bad_input_3():
    with pytest.raises(ValueError, match="size of error does not match size of value"):
        NormalConstraint(["a", "b"], [1, 2], np.eye(3))


def test_NormalConstraint_bad_input_4():
    with pytest.raises(ValueError, match="positive definite"):
        NormalConstraint(["a", "b"], [1, 2], [[1, 1], [1, 1]])


def test_NormalConstraint_bad_input_5():
    n = NormalConstraint(["a", "b"], [1, 2], [[1, 0], [0, 1]])

    with pytest.raises(ValueError, match="positive definite"):
        n.covariance = [[1, 1], [1, 1]]


def test_NormalConstraint_bad_input_6():
    with pytest.raises(ValueError, match="cannot have more than two dimensions"):
        NormalConstraint(["a", "b"], [1, 2], np.ones((2, 2, 2)))


dtypes_to_test = [np.float32]
if hasattr(np, "float128"):  # not available on all platforms
    dtypes_to_test.append(np.float128)


def test_multinomial_chi2():
    zero = np.array(0)
    one = np.array(1)

    assert multinomial_chi2(zero, zero) == 0
    assert multinomial_chi2(zero, one) == 0
    assert multinomial_chi2(one, zero) == 0
    n = np.array([(0.0, 0.0)])
    assert_allclose(multinomial_chi2(n, zero), 0)
    assert_allclose(multinomial_chi2(n, one), 0)


@pytest.mark.skipif(
    not hasattr(np, "float128"), reason="float128 not available on all platforms"
)
def test_model_float128():
    def model(x, a):
        x = x.astype(np.float128)
        return a + x

    for cost in (
        UnbinnedNLL([1], model),
        ExtendedUnbinnedNLL([1], lambda x, a: (1, model(x, a))),
        BinnedNLL([1], [1, 2], model),
        BinnedNLL([[1, 1]], [1, 2], model),
        ExtendedBinnedNLL([1], [1, 2], model),
        ExtendedBinnedNLL([[1, 1]], [1, 2], model),
        LeastSquares([1.0], [2.0], [3.0], model),
        LeastSquares([1.0], [2.0], [3.0], model, loss="soft_l1"),
    ):
        assert cost(1).dtype == np.float128

        Minuit(cost, a=0).migrad()  # should not raise


def test_model_performance_warning():
    def model(x, a):
        return [0 for xi in x]

    with pytest.warns(PerformanceWarning):
        BinnedNLL([1], [1.0, 2.0], model)(1)

    with pytest.warns(PerformanceWarning):
        ExtendedBinnedNLL([1], [1.0, 2.0], model)(1)

    with pytest.warns(PerformanceWarning):
        UnbinnedNLL([1], model)(1)

    with pytest.warns(PerformanceWarning):
        ExtendedUnbinnedNLL([1], lambda x, a: (1, model(x, a)))(1)


@pytest.mark.parametrize("cls", (BinnedNLL, ExtendedBinnedNLL))
def test_update_data_with_mask(cls):
    xe = np.arange(0, 4)
    nx = np.diff(expon_cdf(xe, 1))
    nx[0] += 1
    c = cls(nx.copy(), xe, expon_cdf)

    c.mask = [False, True, True]
    assert c(1) == 0
    nx[0] += 1
    c.n = nx
    assert c(1) == 0
    nx[1] += 1
    c.n = nx
    assert c(1) != 0
    nx[0] -= 2
    c.mask = (True, False, True)
    c.n = nx
    assert c(1) == 0


@pytest.mark.parametrize("method", ("jsc", "asy", "da"))
def test_Template(method):
    if method == "asy":
        pytest.importorskip("scipy.special")

    xe = np.array([0, 1, 2, 3])
    t = np.array([[1, 1, 0], [0, 1, 3]])
    n = t[0] + t[1]

    c = Template(n, xe, t, method=method)
    m = Minuit(c, 1, 1)
    assert m.limits[0] == (0, np.inf)
    assert m.limits[1] == (0, np.inf)
    m.migrad()
    assert m.valid
    assert m.ndof == 1
    if method == "asy":
        assert c.errordef == 0.5
        assert_equal(m.fmin.reduced_chi2, np.nan)
        # asy produces values far away from truth in this case
        assert_allclose(m.values, [1, 3], atol=0.2)
    else:
        assert c.errordef == 1.0
        assert_allclose(m.fval, 0, atol=1e-4)
        assert_allclose(m.fmin.reduced_chi2, 0, atol=1e-5)
        assert_allclose(m.values, [2, 4], atol=1e-2)


@pytest.mark.parametrize(
    "template",
    (
        (np.array([1.0, 1.0, 0.0]), np.array([0.0, 1.0, 3.0])),
        (
            np.array([[1.0, 1.0], [1.0, 1.0], [0.0, 0.0]]),
            np.array([[0.0, 0.0], [1.0, 1.0], [3.0, 3.0]]),
        ),
    ),
)
def test_Template_does_not_modify_inputs(template):
    from copy import deepcopy

    xe = np.array([0, 1, 2, 3])
    template_copy = deepcopy(template)
    n = template[0] + template[1]

    Template(n, xe, template, method="da")
    assert_equal(template, template_copy)


def generate(rng, nmc, truth, bins, tf=1, df=1):
    stats = pytest.importorskip("scipy.stats")
    xe = np.linspace(0, 2, bins + 1)
    b = np.diff(stats.truncexpon(1, 0, 2).cdf(xe))
    s = np.diff(stats.norm(1, 0.1).cdf(xe))
    n = b * truth[0] + s * truth[1]
    t = b * nmc, s * nmc
    if rng is not None:
        n = rng.poisson(n / df) * df
        if df != 1:
            n = np.transpose((n, n * df))
        t = [rng.poisson(ti / tf) * tf for ti in t]
        if tf != 1:
            t = [np.transpose((tj, tj * tf)) for tj in t]
    return n, xe, np.array(t)


@pytest.mark.parametrize("method", ("jsc", "asy", "da"))
@pytest.mark.parametrize("with_mask", (False, True))
@pytest.mark.parametrize("weighted_data", (False, True))
def test_Template_weighted(method, with_mask, weighted_data):
    rng = np.random.default_rng(1)
    truth = 750, 250
    z = []
    rng = np.random.default_rng(1)
    for itoy in range(100):
        ni, xe, ti = generate(rng, 400, truth, 15, 1.5, 1.5 if weighted_data else 1)
        c = Template(ni, xe, ti, method=method)
        if with_mask:
            cx = 0.5 * (xe[1:] + xe[:-1])
            c.mask = cx != 1.5
        m = Minuit(c, *truth)
        m.limits = (0, None)
        m.strategy = 0
        for iter in range(10):
            m.migrad(iterate=1)
            m.hesse()
            if m.valid and m.accurate:
                break
        assert m.valid
        z.append((m.values[1] - truth[1]) / m.errors[1])
    assert_allclose(np.mean(z), 0, atol=0.3)
    assert_allclose(np.std(z), 1, rtol=0.1)


def test_Template_bad_input():
    with pytest.raises(ValueError):
        Template([1, 2], [1, 2, 3], [])

    with pytest.raises(ValueError, match="do not match"):
        Template([1, 2], [1, 2, 3], [[1, 2, 3], [1, 2, 3]])

    with pytest.raises(ValueError, match="do not match"):
        Template(
            [1, 2],
            [1, 2, 3],
            [[[1, 2], [3, 4]], [[1, 2], [3, 4], [5, 6]]],
        )

    with pytest.raises(ValueError, match="not understood"):
        Template([1], [1, 2], [[1]], method="foo")

    with pytest.raises(ValueError, match="number of names"):
        Template([1], [1, 2], [[1]], name=("b", "s"))

    with pytest.raises(ValueError, match="model_or_template"):
        Template([1], [1, 2], [[1], None])


def test_Template_visualize():
    pytest.importorskip("matplotlib")

    xe = [0, 1, 2]
    n = [1, 2]
    t = [[1, 2], [5, 4]]

    c = Template(n, xe, t)

    c.visualize((1, 2))

    c.mask = np.array([False, True])
    c.visualize((1, 2))


def test_Template_visualize_2D():
    pytest.importorskip("matplotlib")

    xe = ([0, 1, 2], [0, 1, 2])
    n = [[1, 2], [3, 4]]
    t = [[[1, 2], [1, 2]], [[5, 4], [5, 4]]]

    c = Template(n, xe, t)
    c.visualize((1, 2))


def test_Template_pickle():
    n = np.array([1, 2, 3])
    xe = np.array([0, 1, 2, 3])
    t = np.array([[1, 1, 0], [0, 1, 3]])

    c = Template(n, xe, t)
    b = pickle.dumps(c)
    c2 = pickle.loads(b)

    assert_equal(c.data, c2.data)


def test_Template_with_model():
    n = np.array([3, 2, 3])
    xe = np.array([0, 1, 2, 3])
    t = np.array([0.1, 0, 1])

    c = Template(n, xe, (t, scaled_cdf))
    assert describe(c) == ["x0", "x1_n", "x1_mu", "x1_sigma"]

    m = Minuit(c, 1, 1, 0.5, 1)
    m.limits["x0", "x1_n"] = (0, None)
    m.limits["x1_sigma"] = (0.1, None)
    m.limits["x1_mu"] = (xe[0], xe[-1])
    m.migrad()
    assert m.valid

    c2 = Template(n, xe, (t, scaled_cdf), name=("a", "b", "c", "d"))
    assert describe(c2) == ["a", "b", "c", "d"]
    m = Minuit(c2, 1, 1, 0.5, 1)
    m.limits["d"] = (0.1, None)
    m.migrad()
    assert m.valid


def test_Template_with_model_2D():
    truth1 = (1.0, 0.1, 0.2, 0.3, 0.4, 0.5)
    x1, y1 = mvnorm(*truth1[1:]).rvs(size=int(truth1[0] * 1000), random_state=1).T
    truth2 = (1.0, 0.2, 0.1, 0.4, 0.3, 0.0)
    x2, y2 = mvnorm(*truth2[1:]).rvs(size=int(truth2[0] * 1000), random_state=1).T

    x = np.append(x1, x2)
    y = np.append(y1, y2)
    w, xe, ye = np.histogram2d(x, y, bins=(3, 5))

    def model(xy, n, mux, muy, sx, sy, rho):
        return n * 1000 * mvnorm(mux, muy, sx, sy, rho).cdf(np.transpose(xy))

    x3, y3 = mvnorm(*truth2[1:]).rvs(size=int(truth2[0] * 10000), random_state=2).T
    template = np.histogram2d(x3, y3, bins=(xe, ye))[0]

    cost = Template(w, (xe, ye), (model, template))
    assert cost.ndata == np.prod(w.shape)
    m = Minuit(cost, *truth1, 1)
    m.limits["x0_n", "x0_sx", "x0_sy"] = (0, None)
    m.limits["x0_rho"] = (-1, 1)
    m.migrad()
    assert m.valid
    assert_allclose(m.values, truth1 + (1e3,), rtol=0.1)


def test_Template_with_only_models():
    n = np.array([3, 2, 3])
    xe = np.array([0, 1, 2, 3])

    c = Template(n, xe, (lambda x, n, mu: n * expon_cdf(x, mu), scaled_cdf))
    assert describe(c) == ["x0_n", "x0_mu", "x1_n", "x1_mu", "x1_sigma"]

    m = Minuit(c, 1, 0.5, 1, 0.5, 1)
    m.limits["x0_n", "x1_n", "x1_sigma"] = (0, None)
    m.limits["x0_mu", "x1_mu"] = (xe[0], xe[-1])
    m.migrad()
    assert m.valid


def test_Template_pulls():
    rng = np.random.default_rng(1)
    xe = np.linspace(0, 1, 200)
    tmu = np.array(
        [
            np.diff(norm_cdf(xe, 0.5, 0.1)) * 1000,
            np.diff(expon_cdf(xe, 1)) * 10000,
        ]
    )
    truth = (1000, 2000)
    mu = sum(truth[i] * (tmu[i] / np.sum(tmu[i])) for i in range(2))
    n = rng.poisson(mu)
    t = rng.poisson(tmu)
    c = Template(n, xe, t)
    m = Minuit(c, *truth)
    m.migrad()
    assert m.valid
    pulls = c.pulls(m.values)
    assert np.nanmean(pulls) == pytest.approx(0, abs=0.1)
    assert np.nanvar(pulls) == pytest.approx(1, abs=0.1)


def test_deprecated():
    from iminuit import cost

    with pytest.warns(FutureWarning):
        from iminuit.cost import BarlowBeestonLite
    assert BarlowBeestonLite is cost.Template

    with pytest.warns(FutureWarning):
        from iminuit.cost import barlow_beeston_lite_chi2_jsc
    assert barlow_beeston_lite_chi2_jsc is cost.template_chi2_jsc

    with pytest.warns(FutureWarning):
        from iminuit.cost import barlow_beeston_lite_chi2_hpd
    assert barlow_beeston_lite_chi2_hpd is cost.template_chi2_da


def test_deprecated_Template_method():
    from iminuit import cost

    with pytest.warns(FutureWarning):
        t = Template([1], [2, 3], [[1], [2]], method="hpd")
        t._impl is cost.template_chi2_da


@pytest.mark.parametrize("cost", (BinnedNLL, ExtendedBinnedNLL))
def test_binned_cost_with_model_shape_error_message_1D(cost):
    n = [1, 2]
    edges = [0.1, 0.2, 0.3]

    # cdf should return same length as xe, not same length as n
    def cdf(xe, a):
        return xe[:-1]

    c = cost(n, edges, cdf)
    with pytest.raises(
        ValueError,
        match=(r"output of model has shape \(2,\), but \(3,\) is required"),
    ):
        c(1)


@pytest.mark.parametrize("cost", (BinnedNLL, ExtendedBinnedNLL))
def test_binned_cost_with_model_shape_error_message_2D(cost):
    n = [[1, 2, 3], [4, 5, 6]]
    edges = [0.5, 0.6, 0.7], [0.1, 0.2, 0.3, 0.4]

    def cdf(xye, a):
        # xye has shape (2, N), returned array should be (N,)
        return np.ones(xye.shape[1] - 1)

    c = cost(n, edges, cdf)
    with pytest.raises(
        ValueError,
        match=(r"output of model has shape \(11,\), but \(12,\) is required"),
    ):
        c(1)



from iminuit.util import describe, make_func_code
from iminuit.typing import Annotated, Gt, Lt, Ge, Le, Interval
from math import ldexp
import platform
from functools import wraps
import pytest
import numpy as np
from numpy.typing import NDArray

is_pypy = platform.python_implementation() == "PyPy"


def test_function():
    def f(x, y):
        pass

    assert describe(f) == ["x", "y"]


def test_class_method():
    class A:
        def f(self, x, y):
            pass

    assert describe(A().f) == ["x", "y"]


def test_class_unbound_method():
    class A:
        def f(self, x, y):
            pass

    assert describe(A.f) == ["self", "x", "y"]


def test_functor():
    class A:
        def __call__(self, x, y):
            pass

    assert describe(A()) == ["x", "y"]


def test_builtin_by_parsing_doc():
    assert describe(ldexp) == ["x", "i"]


def test_lambda():
    assert describe(lambda a, b: 0) == ["a", "b"]


def test_generic_function():
    def f(*args):
        pass

    assert describe(f) == []


def test_generic_partial():
    from functools import partial

    def f(*args):
        pass

    partial_f = partial(f, 42, 12, 4)
    assert describe(partial_f) == []


def test_generic_lambda():
    assert describe(lambda *args: 0) == []


def test_generic_class_method():
    class A:
        def f(self, *args):
            pass

    assert describe(A().f) == []


def test_generic_functor():
    class A:
        def __call__(self, *args):
            pass

    assert describe(A()) == []


def test_generic_functor_with_fake_func():
    class A:
        def __init__(self):
            self.func_code = make_func_code(["x", "y"])

        def __call__(self, *args):
            pass

    with pytest.warns(FutureWarning):
        assert describe(A()) == ["x", "y"]


def test_decorated_function():
    def dummy_decorator(f):
        @wraps(f)
        def decorated(*args, **kwargs):
            return f(*args, **kwargs)

        return decorated

    @dummy_decorator
    def one_arg(x):
        pass

    @dummy_decorator
    def many_arg(x, y, z, t):
        pass

    @dummy_decorator
    def kw_only(x, *, y, z):
        pass

    assert describe(one_arg) == list("x")
    assert describe(many_arg) == list("xyzt")
    assert describe(kw_only) == list("xyz")


def test_ambiguous_1():
    def f(x, *args):
        pass

    assert describe(f) == ["x"]


def test_ambiguous_2():
    def f(x, kw=None):
        pass

    assert describe(f) == ["x", "kw"]


def test_ambiguous_3():
    def f(x, **kw):
        pass

    assert describe(f) == ["x"]


def test_ambiguous_4():
    class A:
        def __call__(self, x, **kw):
            pass

    assert describe(A()) == ["x"]


def test_from_docstring_1():
    def f(*args):
        """f(x, y, z)"""

    assert describe(f) == ["x", "y", "z"]


def test_from_docstring_2():
    class Foo:
        def bar(self, *args):
            """Foo.bar(self, int ncall_me =10000, [resume=True, int nsplit=1])"""
            pass

        def baz(self, *args):
            """Foo.baz(self: Foo, ncall_me: int =10000, arg: np.ndarray = [])"""

    assert describe(Foo().bar) == ["ncall_me", "resume", "nsplit"]
    assert describe(Foo().baz) == ["ncall_me", "arg"]


def test_from_docstring_3():
    assert describe(min) == ["iterable", "default", "key"]


def test_from_docstring_4():
    def f(*args):
        """f(a=(), b=[], *args, *[, foo=1])"""

    assert describe(f) == ["a", "b"]


def test_from_bad_docstring_2():
    def foo(*args):
        """foo is some function"""
        pass

    assert describe(foo) == []


def test_with_type_hints():
    def foo(
        x: NDArray,
        a: Annotated[float, Gt(0), Lt(1)],
        b: float,
        c: Annotated[float, 0:],
        d: Annotated[float, Ge(1)],
        e: Annotated[float, Le(2)],
        f: Annotated[float, Interval(gt=2, lt=3)],
        g: Annotated[float, (4, 5)],
    ): ...

    r = describe(foo, annotations=True)
    assert r == {
        "x": None,
        "a": (0, 1),
        "b": None,
        "c": (0, np.inf),
        "d": (1, np.inf),
        "e": (-np.inf, 2),
        "f": (2, 3),
        "g": (4, 5),
    }

    class Foo:
        def __call__(self, x: NDArray, a: Annotated[float, Gt(0), Lt(1)], b: float): ...

    r = describe(Foo.__call__, annotations=True)
    assert r == {"self": None, "x": None, "a": (0, 1), "b": None}

    r = describe(Foo(), annotations=True)
    assert r == {"x": None, "a": (0, 1), "b": None}


def test_with_pydantic_types():
    tp = pytest.importorskip("pydantic.types")

    def foo(
        x: NDArray,
        a: tp.PositiveFloat,
        b: tp.NonNegativeFloat,
        c: float,
        d: Annotated[float, tp.annotated_types.Gt(1)],
        e: Annotated[float, tp.annotated_types.Interval(gt=0, lt=1)],
    ): ...

    r = describe(foo, annotations=True)
    assert r == {
        "x": None,
        "a": (0, np.inf),
        "b": (0, np.inf),
        "c": None,
        "d": (1, np.inf),
        "e": (0, 1),
    }


def test_with_annotated_types():
    tp = pytest.importorskip("annotated_types")

    def foo(
        x: NDArray,
        a: float,
        b: Annotated[float, tp.Gt(1)],
        c: Annotated[float, tp.Interval(gt=0, lt=1)],
    ): ...

    r = describe(foo, annotations=True)
    assert r == {
        "x": None,
        "a": None,
        "b": (1, np.inf),
        "c": (0, 1),
    }


class Foo(float):
    pass


def test_string_annotation_1():
    def f(x, mu: "Foo"):
        pass

    assert describe(f, annotations=True) == {"x": None, "mu": None}


def test_string_annotation_2():
    def f(x, mu: "Annotated[float, Gt(1)]"):
        pass

    assert describe(f, annotations=True) == {"x": None, "mu": (1, np.inf)}


def test_string_annotation_3():
    def f(x, mu: "Bar"):  # noqa
        pass

    assert describe(f, annotations=True) == {"x": None, "mu": None}



from iminuit._hide_modules import hide_modules
from iminuit.cost import LeastSquares
import pytest

pytest.importorskip("ipywidgets")


def test_interactive():
    pytest.importorskip("matplotlib")
    import iminuit

    cost = LeastSquares([1.1, 2.2], [3.3, 4.4], 1, lambda x, a: a * x)

    iminuit.Minuit(cost, 1).interactive()

    with hide_modules("ipywidgets", reload="iminuit"):
        with pytest.raises(ModuleNotFoundError, match="Please install"):
            iminuit.Minuit(cost, 1).interactive()



from iminuit._hide_modules import hide_modules
import pytest
from iminuit.warnings import OptionalDependencyWarning


def cost(x, y):
    return x**2 + y**2


def test_mncontour_cl():
    with hide_modules("scipy", reload="iminuit"):
        import iminuit

        m = iminuit.Minuit(cost, 1, 1)
        m.migrad()
        m.mncontour("x", "y")
        for cl in (0.68, 0.9, 0.95, 0.99, 1, 2, 3, 4, 5):
            m.mncontour("x", "y", cl=cl)

        with pytest.raises(ImportError):
            m.mncontour("x", "y", cl=0.1)


def test_mncontour_interpolated():
    with hide_modules("scipy", reload="iminuit"):
        import iminuit

        with pytest.warns(
            OptionalDependencyWarning,
            match="interpolation requires optional package 'scipy'",
        ):
            m = iminuit.Minuit(cost, 1, 1)
            m.migrad()
            pts = m.mncontour("x", "y", size=20, interpolated=200)
            assert len(pts) == 21


def test_minos_cl():
    with hide_modules("scipy"):
        import iminuit

        m = iminuit.Minuit(cost, 1, 1)
        m.migrad()
        m.minos()
        for cl in (0.68, 0.9, 0.95, 0.99, 1, 2, 3, 4, 5):
            m.minos(cl=cl)

        with pytest.raises(ModuleNotFoundError):
            m.minos(cl=0.1)


def test_missing_scipy():
    with hide_modules("scipy"):
        import iminuit

        m = iminuit.Minuit(cost, 1, 1)
        with pytest.raises(ModuleNotFoundError):
            m.scipy()



from iminuit import util
import pytest
from argparse import Namespace
from numpy.testing import assert_equal, assert_allclose
import numpy as np
from iminuit._core import MnUserParameterState
from iminuit._optional_dependencies import optional_module_for
import pickle
from iminuit._hide_modules import hide_modules

try:
    import scipy  # noqa

    scipy_available = True
except ModuleNotFoundError:
    scipy_available = False


def test_ndim():
    ndim = util._ndim
    assert ndim(1) == 0
    assert ndim([]) == 1
    assert ndim([[]]) == 2
    assert ndim(None) == 0
    assert ndim((None, None)) == 1
    assert ndim(((1, 2), None)) == 2
    assert ndim((None, (1, 2))) == 2


def test_BasicView():
    with pytest.raises(TypeError):
        util.BasicView(None, 2)


def test_ValueView():
    state = MnUserParameterState()
    state.add("x", 1.0, 0.1)
    state.add("y", 2.2, 0.1)
    state.add("z", 3.3, 0.1)

    v = util.ValueView(
        Namespace(
            _var2pos={"x": 0, "y": 1, "z": 2},
            _pos2var=("x", "y", "z"),
            npar=3,
            _last_state=state,
            _copy_state_if_needed=lambda: None,
        )
    )

    assert v == v
    assert v == (1.0, 2.2, 3.3)
    assert v != (1.0, 2.1, 3.3)
    assert v != 0

    assert repr(v) == "<ValueView x=1.0 y=2.2 z=3.3>"
    assert str(v) == repr(v)

    v[:] = (1, 2, 3)

    assert_equal(v[:3], (1, 2, 3))
    assert_equal(v[0:3], (1, 2, 3))
    assert_equal(v[0:10], (1, 2, 3))

    assert_equal(v, (1, 2, 3))
    v[1:] = 4
    assert_equal(v, (1, 4, 4))
    v["y"] = 2
    assert_equal(v, (1, 2, 4))
    v["y":] = 3
    assert_equal(v, (1, 3, 3))
    v[:"z"] = 2
    assert_equal(v, (2, 2, 3))

    v_dict = v.to_dict()
    assert isinstance(v_dict, dict)
    assert v_dict["x"] == v["x"]
    assert v_dict["y"] == v["y"]
    assert v_dict["z"] == v["z"]

    v[:] = (1, 2, 3)
    assert_equal(v[["x", "z"]], (1, 3))
    assert_equal(v[[2, 0]], (3, 1))
    v[["x", "z"]] = (3, 1)
    assert_equal(v, (3, 2, 1))


def test_FixedView_as_mask_for_other_views():
    state = MnUserParameterState()
    state.add("x", 1, 0.1)
    state.add("y", 2, 0.1)
    state.add("z", 3, 0.1)

    fake_minuit = Namespace(
        _var2pos={"x": 0, "y": 1, "z": 2},
        _pos2var=("x", "y", "z"),
        npar=3,
        _last_state=state,
        _copy_state_if_needed=lambda: None,
    )

    v = util.ValueView(fake_minuit)
    f = util.FixedView(fake_minuit)
    f[1] = True

    assert_equal(f, [False, True, False])
    assert_equal(v[f], [2])
    assert_equal(v[~f], [1, 3])
    v[f] = 5
    assert_equal(v, [1, 5, 3])
    v[~f] = [2, 4]
    assert_equal(v, [2, 5, 4])


def test_FixedView_comparison_with_broadcasting():
    state = MnUserParameterState()
    state.add("x", 1, 0.1)
    state.add("y", 2, 0.1)
    state.add("z", 3, 0.1)

    fake_minuit = Namespace(
        _var2pos={"x": 0, "y": 1, "z": 2},
        _pos2var=("x", "y", "z"),
        npar=3,
        _last_state=state,
        _copy_state_if_needed=lambda: None,
    )

    f = util.FixedView(fake_minuit)

    assert_equal(f, [False, False, False])

    # broadcasting
    assert f == False  # noqa
    f[0] = True
    assert_equal(f, [True, False, False])
    assert f != False  # noqa


def test_Matrix():
    m = util.Matrix(("a", "b"))
    m[:] = [[1, 2], [2, 8]]
    assert_equal(m, ((1, 2), (2, 8)))
    assert repr(m) == "[[1. 2.]\n [2. 8.]]"
    c = m.correlation()
    assert_allclose(c, ((1.0, 0.5**0.5), (0.5**0.5, 1.0)))
    assert m["a", "b"] == 2.0
    assert m["a", 1] == 2.0
    assert m[1, "a"] == 2.0
    assert m[1, 1] == 8.0
    assert_equal(m[0], [1, 2])
    assert_equal(m["b"], (2, 8))
    assert_equal(m[:], [[1, 2], [2, 8]])
    assert_equal(m[:, 0], [1, 2])
    assert_equal(m[:, 1], [2, 8])

    # this swaps rows and cols
    assert_equal(m[[1, 0]], [[8, 2], [2, 1]])

    m *= 2
    assert_equal(m, ((2, 4), (4, 16)))

    m2 = np.dot(m, (1, 1))
    assert repr(m2) == "[ 6. 20.]"
    assert str(m2) == "[ 6. 20.]"
    assert_allclose(m2, (6, 20))

    # matrix is always square

    m = util.Matrix(("a", "b", "c"))
    m[:] = np.arange(9).reshape((3, 3))
    # [0 1 2
    #  3 4 5
    #  6 7 8]

    # m1 = m[:2]
    # assert_equal(m1, [[0, 1], [3, 4]])
    m2 = m[[0, 2]]
    assert_equal(m2, [[0, 2], [6, 8]])
    m3 = m[["a", "c"]]
    assert_equal(m3, [[0, 2], [6, 8]])

    d = m.to_dict()
    assert list(d.keys()) == [
        ("a", "a"),
        ("a", "b"),
        ("a", "c"),
        ("b", "b"),
        ("b", "c"),
        ("c", "c"),
    ]
    for k, v in d.items():
        assert v == m[k]

    with pytest.raises(TypeError):
        util.Matrix("ab")

    with pytest.raises(TypeError):
        util.Matrix(1)

    m2 = pickle.loads(pickle.dumps(m))
    assert type(m2) is util.Matrix

    assert_equal(m2, m)
    assert m2._var2pos == m._var2pos


def test_Param():
    values = 3, "foo", 1.2, 3.4, None, False, False, 42, None
    p = util.Param(*values)

    assert p.number == 3
    assert p.name == "foo"
    assert p.value == 1.2
    assert p.error == 3.4
    assert p.merror is None
    assert not p.is_const
    assert not p.is_fixed
    assert p.has_limits
    assert p.has_lower_limit
    assert not p.has_upper_limit
    assert p.lower_limit == 42
    assert p.upper_limit is None

    assert repr(p) == (
        "Param(number=3, name='foo', value=1.2, error=3.4, merror=None, "
        "is_const=False, is_fixed=False, lower_limit=42, upper_limit=None)"
    )


def test_Params():
    p = util.Params(
        [
            util.Param(0, "foo", 1.2, 3.4, None, False, False, 42, None),
            util.Param(1, "bar", 3.4, 4.5, None, False, False, 42, None),
        ]
    )

    assert repr(p) == repr((p[0], p[1]))
    assert p[0].number == 0
    assert p[1].number == 1
    assert p["foo"].number == 0
    assert p["bar"].number == 1


def test_MError():
    me = util.MError(
        1,
        "x",
        0.1,
        0.2,
        True,
        True,
        True,
        False,
        False,
        False,
        False,
        False,
        False,
        11,
        0.7,
    )

    assert repr(me) == (
        "<MError number=1 name='x' lower=0.1 upper=0.2 is_valid=True lower_valid=True "
        "upper_valid=True at_lower_limit=False at_upper_limit=False "
        "at_lower_max_fcn=False at_upper_max_fcn=False lower_new_min=False "
        "upper_new_min=False nfcn=11 min=0.7>"
    )

    assert me == util.MError(
        1,
        "x",
        0.1,
        0.2,
        True,
        True,
        True,
        False,
        False,
        False,
        False,
        False,
        False,
        11,
        0.7,
    )

    assert me != util.MError(
        1,
        "x",
        0.1,
        0.2,
        True,
        True,
        True,
        False,
        False,
        False,
        False,
        False,
        False,
        11,
        0.8,
    )


def test_MErrors():
    mes = util.MErrors(
        x=util.MError(
            1,
            "x",
            0.1,
            0.2,
            True,
            True,
            True,
            False,
            False,
            False,
            False,
            False,
            False,
            11,
            0.7,
        )
    )

    assert repr(mes) == f"<MErrors\n  {mes['x']!r}\n>"


@pytest.mark.parametrize("errordef", (0.5, 1.0))
def test_FMin(errordef):
    fm = Namespace(
        fval=1.23456e-10,
        edm=1.23456e-10,
        errordef=errordef,
        is_valid=True,
        has_valid_parameters=True,
        has_accurate_covar=True,
        has_posdef_covar=True,
        has_made_posdef_covar=False,
        hesse_failed=False,
        has_covariance=True,
        is_above_max_edm=False,
        has_reached_call_limit=False,
        has_parameters_at_limit=False,
        state=[],
    )
    fmin = util.FMin(fm, "foo", 1, 2, 1, 0.1, 1.2)
    assert {x for x in dir(fmin) if not x.startswith("_")} == {
        "algorithm",
        "edm",
        "edm_goal",
        "errordef",
        "fval",
        "reduced_chi2",
        "nfcn",
        "ngrad",
        "is_valid",
        "has_accurate_covar",
        "has_valid_parameters",
        "has_posdef_covar",
        "has_made_posdef_covar",
        "hesse_failed",
        "has_covariance",
        "is_above_max_edm",
        "has_reached_call_limit",
        "has_parameters_at_limit",
        "time",
    }
    assert fmin.algorithm == "foo"
    assert fmin.edm == 1.23456e-10
    assert fmin.edm_goal == 0.1
    assert not fmin.has_parameters_at_limit
    assert fmin.time == 1.2

    assert fmin == util.FMin(fm, "foo", 1, 2, 1, 0.1, 1.2)
    assert fmin != util.FMin(fm, "foo", 1, 2, 1, 0.3, 1.2)
    assert fmin != util.FMin(fm, "bar", 1, 2, 1, 0.1, 1.2)
    assert fmin != util.FMin(fm, "foo", 1, 2, 1, 0.1, 1.5)

    if errordef == 1:
        reduced_chi2 = fmin.fval
    else:
        reduced_chi2 = np.nan

    assert repr(fmin) == (
        f"<FMin algorithm='foo' edm=1.23456e-10 edm_goal=0.1 errordef={errordef}"
        " fval=1.23456e-10"
        " has_accurate_covar=True has_covariance=True has_made_posdef_covar=False"
        " has_parameters_at_limit=False has_posdef_covar=True"
        " has_reached_call_limit=False has_valid_parameters=True"
        " hesse_failed=False is_above_max_edm=False is_valid=True"
        f" nfcn=1 ngrad=2 reduced_chi2={reduced_chi2} time=1.2>"
    )


def test_normalize_limit():
    assert util._normalize_limit(None) == (-np.inf, np.inf)
    assert util._normalize_limit((None, 2)) == (-np.inf, 2)
    assert util._normalize_limit((2, None)) == (2, np.inf)
    assert util._normalize_limit((None, None)) == (-np.inf, np.inf)
    with pytest.raises(ValueError):
        util._normalize_limit((3, 2))


def test_guess_initial_step():
    assert util._guess_initial_step(0) == 0.1
    assert util._guess_initial_step(1) == 0.01


def test_address_of_cfunc():
    nb = pytest.importorskip("numba")

    nb_sig = nb.types.double(nb.types.uintc, nb.types.CPointer(nb.types.double))

    @nb.cfunc(nb_sig)
    def fcn(n, x):
        x = nb.carray(x, (n,))
        r = 0.0
        for i in range(n):
            r += (x[i] - i) ** 2
        return r

    from ctypes import cast, c_void_p, CFUNCTYPE, POINTER, c_double, c_uint32

    address = cast(fcn.ctypes, c_void_p).value
    assert util._address_of_cfunc(fcn) == address

    # let's see if we can call the function pointer, going full circle
    c_sig = CFUNCTYPE(c_double, c_uint32, POINTER(c_double))
    c_fcn = cast(address, c_sig)

    v = np.array((1.0, 2.0))
    assert c_fcn(2, v.ctypes.data_as(POINTER(c_double))) == 2.0


def test_address_of_cfunc_bad_signature():
    nb = pytest.importorskip("numba")

    nb_sig = nb.types.double(nb.types.double, nb.types.CPointer(nb.types.double))

    @nb.cfunc(nb_sig)
    def fcn(y, x):
        return 0

    assert util._address_of_cfunc(fcn) == 0


def test_make_func_code():
    with pytest.warns(FutureWarning):
        fc = util.make_func_code(["a", "b"])
    assert fc.co_varnames == ("a", "b")
    assert fc.co_argcount == 2

    with pytest.warns(FutureWarning):
        fc = util.make_func_code(("x",))
    assert fc.co_varnames == ("x",)
    assert fc.co_argcount == 1


def test_make_with_signature():
    def f(a, b):
        return a + b

    f1 = util.make_with_signature(f, "x", "y")
    assert util.describe(f1) == ["x", "y"]
    assert f1(1, 2) == f(1, 2)
    f2 = util.make_with_signature(f, b="z")
    assert util.describe(f2) == ["a", "z"]
    assert f2(1, 2) == f(1, 2)
    assert f1 is not f2
    f3 = util.make_with_signature(f, "x", b="z")
    assert util.describe(f3) == ["x", "z"]
    assert f3(1, 2) == f(1, 2)

    # check that arguments are not overridden
    assert util.describe(f1) == ["x", "y"]
    assert util.describe(f) == ["a", "b"]

    with pytest.raises(ValueError):
        util.make_with_signature(f, "a", "b", "c")

    with pytest.raises(ValueError):
        util.make_with_signature(f, "a", "b", "c", b="z")


def test_make_with_signature_on_func_without_code_object():
    class Fcn:
        def __call__(self, x, y):
            return x + y

    f = Fcn()
    assert not hasattr(f, "__code__")

    f1 = util.make_with_signature(f, "x", "y")
    assert util.describe(f1) == ["x", "y"]
    assert f1(1, 2) == f(1, 2)
    assert f1 is not f

    f2 = util.make_with_signature(f1, x="a")
    assert util.describe(f2) == ["a", "y"]
    assert f2(1, 2) == f(1, 2)


def test_merge_signatures():
    def f(x, y, z):
        return x + y + z

    def g(x, a, b):
        return x + a + b

    args, (pf, pg) = util.merge_signatures([f, g])
    assert args == ["x", "y", "z", "a", "b"]
    assert pf == [0, 1, 2]
    assert pg == [0, 3, 4]


@pytest.mark.skipif(not scipy_available, reason="needs scipy")
def test_propagate_1():
    cov = [
        [1.0, 0.1, 0.2],
        [0.1, 2.0, 0.3],
        [0.2, 0.3, 3.0],
    ]
    x = [1, 2, 3]

    def fn(x):
        return 2 * x + 1

    with pytest.warns(FutureWarning):
        y, ycov = util.propagate(fn, x, cov)
    np.testing.assert_allclose(y, [3, 5, 7])
    np.testing.assert_allclose(
        ycov, [[4, 0.4, 0.8], [0.4, 8, 1.2], [0.8, 1.2, 12]], rtol=1e-3
    )

    with pytest.warns(FutureWarning):
        y, ycov = util.propagate(fn, [1], [[2]])
    np.testing.assert_allclose(y, 3)
    np.testing.assert_allclose(ycov, 8, rtol=1e-3)


@pytest.mark.skipif(not scipy_available, reason="needs scipy")
def test_propagate_2():
    cov = [
        [1.0, 0.1, 0.2],
        [0.1, 2.0, 0.3],
        [0.2, 0.3, 3.0],
    ]
    x = [1.0, 2.0, 3.0]

    a = 0.5 * np.arange(30).reshape((10, 3))

    def fn(x):
        return np.dot(a, x)

    with pytest.warns(FutureWarning):
        y, ycov = util.propagate(fn, x, cov)
    np.testing.assert_equal(y, fn(x))
    np.testing.assert_allclose(ycov, np.einsum("ij,kl,jl", a, a, cov), rtol=1e-3)

    def fn(x):
        return np.linalg.multi_dot([x.T, cov, x])

    with pytest.warns(FutureWarning):
        y, ycov = util.propagate(fn, x, cov)
    np.testing.assert_equal(y, fn(np.array(x)))
    jac = 2 * np.dot(cov, x)
    np.testing.assert_allclose(ycov, np.einsum("i,k,ik", jac, jac, cov), rtol=1e-3)


@pytest.mark.skipif(not scipy_available, reason="needs scipy")
def test_propagate_3():
    # matrices with full zero rows and columns are supported
    cov = [
        [1.0, 0.0, 0.2],
        [0.0, 0.0, 0.0],
        [0.2, 0.0, 3.0],
    ]
    x = [1.0, 2.0, 3.0]

    def fn(x):
        return 2 * x + 1

    with pytest.warns(FutureWarning):
        y, ycov = util.propagate(fn, x, cov)
    np.testing.assert_allclose(y, [3, 5, 7])
    np.testing.assert_allclose(ycov, [[4, 0.0, 0.8], [0.0, 0.0, 0.0], [0.8, 0.0, 12]])


@pytest.mark.skipif(not scipy_available, reason="needs scipy")
def test_propagate_on_bad_input():
    cov = [[np.nan, 0.0], [0.0, 1.0]]
    x = [1.0, 2.0]

    def fn(x):
        return 2 * x + 1

    with pytest.warns(FutureWarning):
        with pytest.raises(ValueError):
            util.propagate(fn, x, cov)

    with pytest.warns(FutureWarning):
        with pytest.raises(ValueError):
            util.propagate(fn, x, 1)

    cov = [[1.0], [1.0]]
    with pytest.warns(FutureWarning):
        with pytest.raises(ValueError):
            util.propagate(fn, x, cov)


def test_iterate():
    assert list(util._iterate(1)) == [1]
    assert list(util._iterate([1, 2])) == [1, 2]


def test_replace_none():
    assert util._replace_none(None, 2) == 2
    assert util._replace_none(3, 2) == 3


def test_progressbar(capsys):
    with util.ProgressBar(max_value=4) as bar:
        for i in range(4):
            bar += 1
    stdout, stderr = capsys.readouterr()
    assert stdout == "\r0 %\r25 %\r50 %\r75 %\r100 %\r     "


def test_progressbar_no_ipykernel(capsys):
    with hide_modules("ipykernel"):
        with util.ProgressBar(max_value=4) as bar:
            for i in range(4):
                bar += 1
        stdout, stderr = capsys.readouterr()
        assert stdout == "\r0 %\r25 %\r50 %\r75 %\r100 %\r     "


def test_progressbar_html(capsys):
    import sys

    m_iostream = pytest.importorskip("ipykernel.iostream")
    m_display = pytest.importorskip("IPython.display")

    class OutStream:
        def write(self, s):
            original_stdout.write(s)

        def flush(self):
            original_stdout.flush()

    def display(msg, **kwargs):
        sys.stdout.write(msg._repr_html_())

    original_stdout = sys.stdout
    # make ProgressBar think it is running in Jupyter
    sys.stdout = OutStream()

    # monkey-patching our mockups
    m_iostream.OutStream = OutStream
    m_display.display = display

    with util.ProgressBar(max_value=4) as bar:
        for i in range(4):
            bar += 1

    sys.stdout = original_stdout
    stdout, stderr = capsys.readouterr()

    assert stdout == (
        "<progress value='0' max='100'></progress> 0 %"
        "<progress value='25' max='100'></progress> 25 %"
        "<progress value='50' max='100'></progress> 50 %"
        "<progress value='75' max='100'></progress> 75 %"
        "<progress value='100' max='100'></progress> 100 %"
    )


@pytest.mark.parametrize(
    "mask_expected",
    (
        (None, [[0, 1, 2]]),
        ([True, False], [[0]]),
        ([False, True], [[1]]),
        ([True, True, True, False, True, True, False, True], [[0, 1, 2], [4, 5], [7]]),
    ),
)
def test_histogram_segments(mask_expected):
    mask, expected = mask_expected
    if mask is None:
        xe = np.linspace(0, 1, 4)
        masked = np.arange(3)
    else:
        xe = np.linspace(0, 1, len(mask))
        masked = np.arange(len(mask))[np.array(mask)]
    segments = util._histogram_segments(mask, xe, masked)
    assert_equal([s[0] for s in segments], expected)


@pytest.mark.parametrize(
    "fn_expected", ((lambda x: x, 15), (lambda x: x**11, 40), (np.log, 80))
)
def test_smart_sampling_1(fn_expected):
    fn, expected = fn_expected
    x, y = util._smart_sampling(fn, 1e-10, 5)
    assert len(y) < expected


def test_smart_sampling_2():
    # should not raise a warning
    x, y = util._smart_sampling(np.log, 1e-10, 1, tol=1e-5)
    assert 0 < len(x) < 1000


def test_smart_sampling_3():
    def step(x):
        return np.where(x > 0.5, 0, 1)

    with pytest.warns(RuntimeWarning, match="Iteration limit"):
        x, y = util._smart_sampling(step, 0, 1, tol=0)
    assert 0 < len(x) < 80


def test_smart_sampling_4():
    from time import sleep

    def step(x):
        sleep(0.1)
        return np.where(x > 0.5, 0, 1)

    with pytest.warns(RuntimeWarning, match="Time limit"):
        x, y = util._smart_sampling(step, 0, 1, maxtime=0)
    assert 0 < len(x) < 10


@pytest.mark.parametrize(
    "x,expected",
    [([1, 2, 3], False), ([-1, 1000, 100000], False), ([1, 10, 100], True)],
)
def test_detect_log_spacing_1(x, expected):
    assert util._detect_log_spacing(x) == expected


def test_optional_module_for_1():
    with optional_module_for("foo"):
        import iminuit  # noqa


def test_optional_module_for_2():
    from iminuit.warnings import OptionalDependencyWarning

    with pytest.warns(
        OptionalDependencyWarning, match="foo requires optional package 'foobarbaz'"
    ):
        with optional_module_for("foo"):
            import foobarbaz  # noqa


def test_optional_module_for_3():
    from iminuit.warnings import OptionalDependencyWarning

    with pytest.warns(
        OptionalDependencyWarning, match="foo requires optional package 'foo'"
    ):
        with optional_module_for("foo", replace={"foobarbaz": "foo"}):
            import foobarbaz  # noqa


def test_positive_definite():
    assert util.is_positive_definite([[1, 0], [0, 1]])
    assert not util.is_positive_definite([[1, 1], [1, 1]])
    assert not util.is_positive_definite([[1, 0], [1, 1]])



import os
from iminuit import __version__ as version
import sys

sys.path.append(".")

from root_version import root_version  # noqa

# release and version are special variables used by sphinx

with open("../README.rst") as f:
    readme_content = f.read()

readme_content = readme_content.replace("doc/", "")
readme_content = readme_content.replace(
    ".. version-marker-do-not-remove",
    "**These docs are for iminuit version:** |release|",
)

with open("index.rst.in") as f:
    index_content = f.read()

with open("index.rst", "w") as f:
    f.write(readme_content + index_content)

release = f"{version} compiled with ROOT-{root_version}"

project = "iminuit"
copyright = "2022, Hans Dembinski and the iminuit team"

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    "sphinx.ext.napoleon",
    "sphinx.ext.autodoc",
    "sphinx.ext.mathjax",
    "sphinx.ext.autosummary",
    # 'matplotlib.sphinxext.only_directives',
    "matplotlib.sphinxext.plot_directive",
    "nbsphinx",
    "IPython.sphinxext.ipython_console_highlighting",
]

nbsphinx_kernel_name = "python3"
nbsphinx_execute_arguments = [
    "--InlineBackend.figure_formats={'svg', 'pdf'}",
    "--InlineBackend.rc=figure.dpi=96",
]
nbsphinx_execute = "auto"
# use FAST=1 to speed up doc build
if bool(os.environ.get("FAST", False)):
    nbsphinx_execute = "never"

autoclass_content = "both"
autosummary_generate = True
autodoc_member_order = "groupwise"
autodoc_type_aliases = {"ArrayLike": "ArrayLike"}

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ["_build", "_themes"]

# html_logo = "_static/iminuit_logo.svg"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# http://read-the-docs.readthedocs.org/en/latest/theme.html#how-do-i-use-this-locally-and-on-read-the-docs
# on_rtd is whether we are on readthedocs.org
on_rtd = os.environ.get("READTHEDOCS", None) == "True"
if not on_rtd:
    # Import and set the theme if we're building docs locally
    import sphinx_rtd_theme

    html_theme = "sphinx_rtd_theme"
    html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

nbsphinx_kernel_name = "python3"
nbsphinx_execute_arguments = [
    "--InlineBackend.figure_formats={'svg', 'pdf'}",
    "--InlineBackend.rc=figure.dpi=96",
]

linkcheck_timeout = 10
linkcheck_allow_unauthorized = True
linkcheck_report_timeouts_as_broken = False
linkcheck_ignore = ["https://doi.org/10.2307%2F2347496"]



from iminuit._deprecated import deprecated, deprecated_parameter
import pytest


def test_deprecated_func_1():
    @deprecated("bla")
    def func(x):
        pass

    with pytest.warns(
        FutureWarning,
        match="func is deprecated: bla",
    ):
        func(1)


def test_deprecated_func_2():
    @deprecated("bla", removal="1.0")
    def func(x):
        pass

    with pytest.warns(
        DeprecationWarning,
        match="func is deprecated and will be removed in version 1.0: bla",
    ):
        func(1)


def test_deprecated_func_3():
    @deprecated("bla", removal="1000.0")
    def func(x):
        pass

    with pytest.warns(
        FutureWarning,
        match="func is deprecated and will be removed in version 1000.0: bla",
    ):
        func(1)


def test_deprecated_parameter():
    @deprecated_parameter(foo="bar")
    def some_function(x, y, foo):
        pass

    some_function(1, 2, foo=3)

    with pytest.warns(
        FutureWarning,
        match="keyword 'bar' is deprecated, please use 'foo'",
    ):
        some_function(1, 2, bar=3)

    with pytest.raises(TypeError):
        some_function(x=1, baz=3, y=2)



import numpy as np
from numpy.random import default_rng
from matplotlib import pyplot as plt
import matplotlib as mpl
from matplotlib.ticker import LogLocator
import os
import pickle

mpl.rcParams.update(mpl.rcParamsDefault)


class TrackingFcn:
    errordef = 1

    def __init__(self, rng, npar):
        self.ncall = 0
        self.y = 5 * rng.standard_normal(npar)

    def __call__(self, par, *args):
        self.ncall += 1
        # make problem non-linear
        z = self.y - par
        return np.sum(z**2 + 0.1 * z**4)


class Runner:
    def __init__(self, npars):
        self.npars = npars

    def __call__(self, seed):
        from iminuit import Minuit
        import nlopt
        from scipy.optimize import minimize

        data = []

        rng = default_rng(seed)
        for npar in self.npars:
            fcn = TrackingFcn(rng, npar)
            for stra in (0, 1, 2):
                key = f"Minuit2/strategy={stra}"
                print(key, npar)
                fcn.ncall = 0
                m = Minuit(fcn, np.zeros(npar))
                m.strategy = stra
                m.migrad()
                max_dev = np.max(np.abs(m.np_values() - fcn.y))
                data.append((key, npar, fcn.ncall, max_dev))

            for algo in ("BOBYQA", "NEWUOA", "PRAXIS", "SBPLX"):
                if npar == 1 and algo == "PRAXIS":
                    continue  # PRAXIS does not work for npar==1
                print(algo, npar)
                fcn.ncall = 0
                opt = nlopt.opt(getattr(nlopt, "LN_" + algo), npar)
                opt.set_min_objective(lambda par, grad: fcn(par))
                opt.set_xtol_abs(1e-2)
                try:
                    xopt = opt.optimize(np.zeros(npar))
                    max_dev = np.max(np.abs(xopt - fcn.y))
                    key = f"nlopt/{algo}"
                    data.append((key, npar, fcn.ncall, max_dev))
                except Exception:
                    pass

            for algo in ("BFGS", "CG", "Powell", "Nelder-Mead"):
                print(algo, npar)
                fcn.ncall = 0
                result = minimize(fcn, np.zeros(npar), method=algo, jac=False)
                max_dev = np.max(np.abs(result.x - fcn.y))
                key = f"scipy/{algo}"
                data.append((key, npar, fcn.ncall, max_dev))

        return data


if os.path.exists("bench.pkl"):
    with open("bench.pkl", "rb") as f:
        results = pickle.load(f)
else:
    npars = (1, 2, 3, 4, 6, 10, 20, 30, 40, 60, 100)

    from numpy.random import SeedSequence
    from concurrent.futures import ProcessPoolExecutor as Pool

    sg = SeedSequence(1)
    with Pool() as p:
        results = tuple(p.map(Runner(npars), sg.spawn(16)))

    with open("bench.pkl", "wb") as f:
        pickle.dump(results, f)


# plt.figure()
# f = TrackingFcn(default_rng(), 2)
# x = np.linspace(-10, 10)
# X, Y = np.meshgrid(x, x)
# F = np.empty_like(X)
# for i, xi in enumerate(x):
#     for j, yi in enumerate(x):
#         F[i, j] = f((xi, yi))
# plt.pcolormesh(X, Y, F.T)
# plt.colorbar()

methods = {}
for data in results:
    for key, npar, ncal, maxdev in data:
        methods.setdefault(key, {}).setdefault(npar, []).append((ncal, maxdev))

fig, ax = plt.subplots(1, 2, figsize=(10, 5), sharex=True)
plt.subplots_adjust(
    top=0.96, bottom=0.14, left=0.075, right=0.81, hspace=0.2, wspace=0.25
)
handles = []
labels = []
markers = (
    ("o", 10),
    ("s", 7),
    ("D", 7),
    ("<", 7),
    (">", 7),
    ("^", 7),
    ("v", 7),
    ("*", 9),
    ("X", 7),
    ("P", 7),
    ("p", 8),
)


for method, (m, ms) in zip(sorted(methods), markers):
    ls = "-"
    lw = 1
    zorder = None
    color = None
    mfc = None
    mew = 1
    if "Minuit" in method:
        ls = "-"
        lw = 2
        zorder = 10
        color = "k"
        mfc = "w"
        mew = 2

    data = methods[method]
    npars = np.sort(list(data))
    ncalls = np.empty_like(npars)
    max_devs = np.empty_like(npars, dtype=float)
    for i, npar in enumerate(npars):
        nc, md = np.transpose(data[npar])
        ncalls[i] = np.median(nc)
        max_devs[i] = np.median(md)
    plt.sca(ax[0])
    (p,) = plt.plot(
        npars,
        ncalls / npars,
        ls=ls,
        lw=lw,
        marker=m,
        ms=ms,
        zorder=zorder,
        color=color,
        mfc=mfc,
        mew=mew,
    )
    handles.append(p)
    labels.append(method)
    plt.xlabel("$N_\\mathrm{par}$")
    plt.ylabel("$N_\\mathrm{call}$ / $N_\\mathrm{par}$")
    plt.loglog()
    plt.ylim(8, 5e2)
    plt.xlim(0.7, 150)
    plt.sca(ax[1])
    plt.xlabel("$N_\\mathrm{par}$")
    plt.ylabel("maximum deviation")
    plt.plot(
        npars,
        max_devs,
        lw=lw,
        ls=ls,
        marker=m,
        ms=ms,
        zorder=zorder,
        color=color,
        mfc=mfc,
        mew=mew,
    )
    plt.loglog()
    plt.gca().yaxis.set_major_locator(LogLocator(numticks=100))

plt.figlegend(handles, labels, loc="center right", fontsize="small")
plt.savefig("bench.svg")

plt.figure(constrained_layout=True)
plt.loglog()
for method, (m, ms) in zip(sorted(methods), markers):
    zorder = None
    color = None
    mfc = None
    mew = 1
    if "Minuit" in method:
        zorder = 10
        color = "k"
        mfc = "w"
        mew = 2

    data = methods[method]

    x = []
    y = []
    s = []
    for npar in (2, 10, 100):
        if npar not in data:
            continue
        nc, md = np.transpose(data[npar])
        x.append(np.median(nc) / npar)
        y.append(np.median(md))
        s.append(50 * npar**0.5)

    plt.scatter(x, y, s, marker=m, color=mfc, edgecolor=color, zorder=zorder)

plt.xlabel("$N_\\mathrm{call}$ / $N_\\mathrm{par}$")
plt.ylabel("maximum deviation")
plt.title("small: npar = 2, medium: npar = 10, large: npar = 100")
plt.savefig("bench2d.svg")

plt.show()



from iminuit._parse_version import parse_version
import pytest


@pytest.mark.parametrize(
    "s,ref",
    [
        ("1.2", (1, 2)),
        ("1.2.3", (1, 2, 3)),
        ("1.2a1", (1, 2)),
        ("1.2.3a1", (1, 2, 3)),
        ("1.2.post1", (1, 2)),
        ("1.2.3.post1", (1, 2, 3)),
        ("1.2a1.dev1", (1, 2)),
        ("1.2.3a1.dev1", (1, 2, 3)),
    ],
)
def test_parse_version(s, ref):
    assert parse_version(s) == ref


def test_parse_version_bad():
    with pytest.raises(ValueError):
        parse_version("a.b.c")



from iminuit import Minuit, cost
import numpy as np
from matplotlib import pyplot as plt


# custom visualization; x, y, model are taken from outer scope
def viz(args):
    plt.plot(x, y, "ok")
    xm = np.linspace(x[0], x[-1], 100)
    plt.plot(xm, model(xm, *args))


def model(x, a, b):
    return a + b * x


x = np.array([1, 2, 3, 4, 5])
y = np.array([1.03, 1.58, 2.03, 2.37, 3.09])
c = cost.LeastSquares(x, y, 0.1, model)
m = Minuit(c, 0.5, 0.5)
m.interactive(viz)
# m.interactive() also works and calls LeastSquares.visualize



import requests
import json
import shutil
from urllib.request import urlopen
from pathlib import Path
import zipfile
import tempfile

organization = "scikit-hep"
project = "iminuit"


def get(url):
    return json.loads(requests.get(url).content)


def get_latest_successful_build():
    # get list of all builds
    artifact = get(
        f"https://api.github.com/repos/{organization}/{project}/actions/artifacts"
    )["artifacts"][0]
    return artifact["id"], artifact["updated_at"], artifact["archive_download_url"]


id, date, url = get_latest_successful_build()

# TODO needs authentication to work
print(f"Downloading artifacts for {id} {date}")

dist = Path("dist")
if dist.exists():
    if any(dist.iterdir()):
        raise SystemExit("Error: ./dist directory is not empty; please delete content")
else:
    dist.mkdir()

print(f"Downloading {id}...")
with tempfile.TemporaryFile(mode="w+b") as tmp:
    with urlopen(url) as src:
        shutil.copyfileobj(src, tmp)
    with zipfile.ZipFile(tmp) as zip:
        for zip_name in zip.namelist():
            fn = dist / Path(zip_name).name
            print("... extracting", fn)
            with zip.open(zip_name) as src, open(fn, "wb") as dst:
                shutil.copyfileobj(src, dst)
